from collections import defaultdict

import numpy as np
import scipy.ndimage as ndimage

import basty.utils.misc as misc
from basty.behavior_mapping.behavioral_windows import BehavioralWindows


class PostProcessing:
    @staticmethod
    def postprocess_wrt_durations(labels, marker_labels, duration_threshold):
        dur_bool = duration_threshold < -1
        duration_threshold = np.abs(duration_threshold)
        if duration_threshold > 1:
            intvls = misc.cont_intvls(labels.astype(int))
            processed_labels = np.empty_like(labels)
            for i in range(1, intvls.shape[0]):
                lbl = labels[intvls[i - 1]]
                intvl_start, intvl_end = intvls[i - 1], intvls[i]
                pp_intvl = ((intvl_end - intvl_start) < duration_threshold) != dur_bool
                if pp_intvl and (lbl in marker_labels):
                    processed_labels[intvl_start:intvl_end] = -1
                else:
                    processed_labels[intvl_start:intvl_end] = lbl
        else:
            processed_labels = labels
        return processed_labels

    @staticmethod
    def get_cont_intvsl_dict(labels):
        intvls = misc.cont_intvls(labels)

        cont_intvls_dict = defaultdict(list)

        for i in range(1, intvls.shape[0]):
            lbl = labels[intvls[i - 1]]
            intvl_start, intvl_end = intvls[i - 1], intvls[i]
            cont_intvls_dict[lbl].append((intvl_start, intvl_end))

        return cont_intvls_dict

    @staticmethod
    def compute_window_majority(labels, winsize, wintype="boxcar"):
        if winsize > 1:
            win_positive_counts = BehavioralWindows.get_behavior_counts(
                labels,
                wintype=wintype,
                winsize=winsize,
                stepsize=1,
            )
            labels_postprocessed = np.array(
                [np.argmax(counts) for counts in (win_positive_counts)]
            )
        else:
            labels_postprocessed = labels
        return labels_postprocessed

    @staticmethod
    def median_filter(X, winsize, **kwargs):
        if X.ndim == 2:
            for i in range(X.shape[1]):
                X[:, i] = ndimage.median_filter(X[:, i], size=winsize, **kwargs)
        elif X.ndim == 1:
            X = ndimage.median_filter(X, size=winsize, **kwargs)
        else:
            raise ValueError("Number of dimensions have to be one or two.")
        return X

    @staticmethod
    def running_mean(X, winsize, **kwargs):
        if X.ndim == 2:
            for i in range(X.shape[1]):
                X[:, i] = ndimage.uniform_filter1d(X[:, i], size=winsize, **kwargs)
        elif X.ndim == 1:
            X = ndimage.uniform_filter1d(X, size=winsize, **kwargs)
        else:
            raise ValueError("Number of dimensions have to be one or two.")
        return X



import logging

import numpy as np
from scipy.signal.windows import get_window
from sklearn.feature_extraction.text import TfidfTransformer
from sklearn.preprocessing import MultiLabelBinarizer, normalize

import basty.utils.misc as misc


class BehavioralWindows:
    def __init__(self, fps, win_cfg={}, use_stride_tricks=False):
        self.win_cfg = win_cfg
        self.logger = logging.getLogger("main")

        self.wintype = self.win_cfg.get("wintype", "boxcar")
        self.winsize = int(self.win_cfg.get("winsize", 10) * fps)
        self.stepsize = int(self.win_cfg.get("stepsize", 5) * fps)
        self.method = self.win_cfg.get("method", "tf-idf")
        self.norm = self.win_cfg.get("norm", None)
        self.use_stride_tricks = use_stride_tricks

        assert self.winsize > 1

        if self.stepsize > self.winsize:
            self.logger.warning("Step-size should be smaller than window-size.")

        if self.method == "tf-idf" and self.wintype != "boxcar":
            raise ValueError(
                "The tf-idf statistics can only be used with boxcar window type."
            )
        if self.norm is None and self.method == "tf-idf":
            self.logger.warning(
                "A normalization method should be specifed for tf-idf representations."
                + "Default is 'L1'."
            )
            self.norm = self.win_cfg.get("norm", "l1")

        if self.norm is not None and self.method == "tf-idf":
            self.logger.warning(
                f"Normalization method is ignored for {self.method} representations."
            )
            self.norm = None

    @staticmethod
    def align_behavioral_windows(win_repr, uniq_labels, num_total_uniq):
        assert win_repr.dim > 1
        win_repr_aligned = np.zeros((win_repr.shape[0], num_total_uniq))
        for idx, lbl in sorted(uniq_labels):
            win_repr_aligned[:, lbl] = win_repr[:, idx]
        return win_repr_aligned

    def get_window_representation(self, bhv_seq):
        if self.method == "tf-idf":
            win_repr = self.get_tfidf_vectors(
                bhv_seq,
                winsize=self.winsize,
                stepsize=self.stepsize,
                norm=self.norm,
                use_stride_tricks=self.use_stride_tricks,
            )
        elif self.method == "binary":
            win_repr = self.get_binary_vectors(
                bhv_seq,
                winsize=self.winsize,
                stepsize=self.stepsize,
                use_stride_tricks=self.use_stride_tricks,
            )
        elif self.method == "count" and self.norm is None:
            win_repr = self.get_behavior_counts(
                bhv_seq,
                wintype=self.wintype,
                winsize=self.winsize,
                stepsize=self.stepsize,
                use_stride_tricks=self.use_stride_tricks,
            )
        elif self.method == "count" and self.norm in ["l1", "l2"]:
            win_repr = self.get_count_vectors(
                bhv_seq,
                wintype=self.wintype,
                winsize=self.winsize,
                stepsize=self.stepsize,
                norm=self.norm,
                use_stride_tricks=self.use_stride_tricks,
            )
        elif self.method == "majority":
            win_repr = self.get_majority_labels(
                bhv_seq,
                wintype=self.wintype,
                winsize=self.winsize,
                stepsize=self.stepsize,
                use_stride_tricks=self.use_stride_tricks,
            )
        else:
            raise ValueError(f"Incompatible pair ({self.method},{self.norm}) is given.")
        return win_repr

    @staticmethod
    def _get_bhv_windows(bhv_seq, winsize, stepsize, use_stride_tricks=False):
        assert winsize > 1
        if use_stride_tricks and stepsize == 1:
            from numpy.lib.stride_tricks import sliding_window_view

            bhv_win = sliding_window_view(bhv_seq, winsize)
        else:
            bhv_win = [
                w
                for w in misc.sliding_window(
                    bhv_seq, winsize=winsize, stepsize=stepsize
                )
            ]
        return bhv_win

    @classmethod
    def get_behavior_counts(
        cls, bhv_seq, wintype="boxcar", winsize=60, stepsize=1, use_stride_tricks=False
    ):
        num_bhv = np.max(bhv_seq) + 1

        bhv_win = cls._get_bhv_windows(
            bhv_seq, winsize, stepsize, use_stride_tricks=use_stride_tricks
        )
        window_arr = get_window(wintype, winsize)

        bhv_count = np.zeros((len(bhv_win), num_bhv), dtype=np.float)

        for i, bhv in enumerate(bhv_win):
            for j, bhv_j in enumerate(bhv):
                bhv_count[i, bhv_j] += window_arr[j]
        return bhv_count

    @classmethod
    def get_tfidf_vectors(
        cls, bhv_seq, winsize=60, stepsize=1, norm="l1", use_stride_tricks=False
    ):
        assert norm in ["l1", "l2"]
        bhv_count = cls.get_behavior_counts(
            bhv_seq,
            winsize=winsize,
            stepsize=stepsize,
            use_stride_tricks=use_stride_tricks,
        )
        vectorizer = TfidfTransformer(norm=norm)
        bhv_tfidf = vectorizer.fit_transform(bhv_count).toarray()
        return bhv_tfidf

    @classmethod
    def get_binary_vectors(
        cls, bhv_seq, winsize=60, stepsize=1, use_stride_tricks=False
    ):
        bhv_win = cls._get_bhv_windows(
            bhv_seq, winsize, stepsize, use_stride_tricks=use_stride_tricks
        )
        mlb_vectorizer = MultiLabelBinarizer()
        bhv_binary = mlb_vectorizer.fit_transform(bhv_win)
        return bhv_binary

    @classmethod
    def get_count_vectors(
        cls,
        bhv_seq,
        wintype="boxcar",
        winsize=60,
        stepsize=1,
        norm="l1",
        use_stride_tricks=False,
    ):
        assert norm in ["l1", "l2", "max"]
        bhv_count = cls.get_behavior_counts(
            bhv_seq,
            wintype=wintype,
            winsize=winsize,
            stepsize=stepsize,
            use_stride_tricks=use_stride_tricks,
        )
        bhv_count_n = normalize(bhv_count, norm=norm)
        return bhv_count_n

    @classmethod
    def get_majority_labels(
        cls,
        bhv_seq,
        wintype="boxcar",
        winsize=60,
        stepsize=1,
        ignore_idx=[],
        use_stride_tricks=False,
    ):
        bhv_win = cls._get_bhv_windows(
            bhv_seq, winsize, stepsize, use_stride_tricks=use_stride_tricks
        )

        if ignore_idx:
            bhv_majority = np.array(
                [misc.most_common(bhv_win[i]) for i in range(len(bhv_win))]
            )
        else:
            bhv_count = cls.get_behavior_counts(
                bhv_seq,
                wintype=wintype,
                winsize=winsize,
                stepsize=stepsize,
                use_stride_tricks=use_stride_tricks,
            )
            bhv_count[:, ignore_idx] = -np.inf
            bhv_majority = [np.argmax(bhv_count[i, :]) for i in range(len(bhv_count))]
        return bhv_majority



import logging
import random

import numpy as np
import pandas as pd
from numpy.lib.stride_tricks import sliding_window_view
from scipy.ndimage.filters import uniform_filter1d

import basty.utils.misc as misc


class BodyPose:
    def __init__(self, pose_cfg={}):
        self.pose_cfg = pose_cfg.copy()
        self.logger = logging.getLogger("main")

        self.counterparts = self.pose_cfg.get("counterparts", {})
        self.singles = self.pose_cfg.get("singles", [])
        self.connected_parts = self.pose_cfg.get("connected_parts", [])

        self.groups = self.pose_cfg.get("groups", {})

        self.defined_points = self.pose_cfg.get("defined_points", {})
        self.centerline = self.pose_cfg.get("centerline", [])

        if len(self.centerline) != 2 and len(self.centerline) != 0:
            raise ValueError(
                "A centerline must be defined by two body-parts"
                + f"Given value {self.centerline} is not suitable for definition."
            )

    @staticmethod
    def get_sub_df_pose(df_pose, ind):
        if ind not in ["x", "y"]:
            raise ValueError(f"Given sub-dataframe indicator {ind} is not defined.")
        sub_df = df_pose[df_pose.columns[df_pose.columns.str.endswith(ind)]]
        sub_df.columns = [name[:-2] for name in sub_df.columns]
        return sub_df

    @staticmethod
    def get_sub_df_coord(df_coord, ind):
        if ind not in ["x", "y", "likelihood"]:
            raise ValueError(f"Given sub-dataframe indicator {ind} is not defined.")
        try:
            sub_df = df_coord[df_coord.columns[df_coord.loc[1] == ind]]
            sub_df.columns = sub_df.iloc[0]
            sub_df = sub_df.drop([0, 1]).fillna(0).reset_index(drop=True)
            assert sub_df.shape[1] > 0
        except AssertionError:
            sub_df = df_coord[df_coord.columns[df_coord.loc[2] == ind]]
            sub_df.columns = sub_df.iloc[1]
            sub_df = sub_df.drop([0, 1, 2]).fillna(0).reset_index(drop=True)
        return sub_df

    def split_sub_df_coord(self, df_coord):
        llh = self.get_sub_df_coord(df_coord, "likelihood").astype(float)
        x = self.get_sub_df_coord(df_coord, "x").astype(float)
        y = self.get_sub_df_coord(df_coord, "y").astype(float)
        return llh, x, y

    def make_frames_egocentric(self, df_pose, per_frame=False):
        spine = self.centerline
        eps = np.finfo(np.float32).eps
        if len(spine) == 0:
            raise ValueError(
                "A centerline must be given to construct egencentric frames."
            )

        dfPose_x = self.get_sub_df_pose(df_pose, "x")
        dfPose_y = self.get_sub_df_pose(df_pose, "y")

        dfPose_new_dict = {}

        if per_frame:
            raise NotImplementedError(
                "Constructing egocentric frames per frame is not implemented yet."
            )
        else:
            s1_x = dfPose_x[spine[0]].to_numpy()
            s1_y = dfPose_y[spine[0]].to_numpy()

            s2_x = dfPose_x[spine[1]].to_numpy()
            s2_y = dfPose_y[spine[1]].to_numpy()

            r2_shifted = np.stack((s2_x - s1_x, s2_y - s1_y), axis=1)

            r2_x = np.linalg.norm(r2_shifted, axis=1)
            r2_y = np.zeros(s2_y.shape[0])
            r2 = np.stack((r2_x, r2_y), axis=1)

            def get_rotation_matrix(v1, v2):
                r1, r2 = (v1 / np.linalg.norm(v1)).reshape(2), (
                    v2 / np.linalg.norm(v2)
                ).reshape(2)
                rc = np.cross(r1, r2)
                rd = np.dot(r1, r2)
                rc_norm = np.linalg.norm(rc)

                kmat = np.array([[0, -rc], [rc, 0]])

                rotation_matrix = (
                    np.eye(2)
                    + kmat
                    + kmat.dot(kmat) * ((1 - rd) / (rc_norm + eps) ** 2)
                )

                return rotation_matrix

            egocentric_transform = np.ndarray((r2.shape[0], r2.shape[1], r2.shape[1]))
            for i in range(egocentric_transform.shape[0]):
                egocentric_transform[i] = get_rotation_matrix(
                    r2_shifted[i, :], r2[i, :]
                )

            # rs_pinv = np.linalg.pinv(r2_shifted[:,:,np.newaxis])
            # egocentric_transform = np.matmul(r2[:, :, np.newaxis], rs_pinv)

            for name in dfPose_x.columns:
                r = np.stack(
                    (
                        dfPose_x[name].to_numpy() - s1_x,
                        dfPose_y[name].to_numpy() - s1_y,
                    ),
                    axis=1,
                )
                r_new = np.matmul(egocentric_transform, r[:, :, np.newaxis])

                dfPose_new_dict[name + "_x"] = r_new[:, 0, 0]
                dfPose_new_dict[name + "_y"] = r_new[:, 1, 0]

            # It's very likely for df_pose to contain NaNs at this point.
            # Interpolating NaN's might be a good idea.
            dfEgocentric = pd.DataFrame.from_dict(dfPose_new_dict)
            return dfEgocentric

    def _finalize_orientation(self, orientations, left_llh, right_llh, winsize=3):
        left_llh = uniform_filter1d(left_llh, size=winsize)
        right_llh = uniform_filter1d(right_llh, size=winsize)

        left_idx = np.asarray(left_llh > right_llh).nonzero()[0]
        right_idx = np.asarray(right_llh > left_llh).nonzero()[0]

        orientations["left"].update(left_idx)
        orientations["right"].update(right_idx)
        orientations["idx"].difference_update(left_idx)
        orientations["idx"].difference_update(right_idx)

        for idx in orientations["idx"]:
            lc = np.abs(np.min(np.array(list(orientations["left"])) - idx))
            rc = np.abs(np.min(np.array(list(orientations["right"])) - idx))
            if lc < rc:
                orientations["left"].add(idx)
            elif rc < lc:
                orientations["right"].add(idx)
            else:
                rand_orient = random.choice(["right", "left"])
                orientations[rand_orient].add(idx)

        orientations["idx"].difference_update(orientations["idx"])

        return orientations

    def _window_count_orientation(self, orientations, left_llh, right_llh, winsize):
        likely_left = left_llh > right_llh

        for idx, window in enumerate(sliding_window_view(likely_left, 2 * winsize + 1)):
            if idx in orientations["idx"]:
                left_count = np.count_nonzero(window)
                right_count = winsize - left_count
                if left_count > winsize:
                    orientations["left"].add(idx)
                    orientations["idx"].remove(idx)
                elif right_count > winsize:
                    orientations["right"].add(idx)
                    orientations["idx"].remove(idx)
                else:
                    pass

        return orientations

    def _nearest_time_point_orientation(self, orientations, left_llh, right_llh):
        for idx in list(orientations["idx"]):
            left_closest = (
                np.min(orientations["left"], key=lambda x: np.abs(x - idx))
                if len(orientations["left"]) > 0
                else -1
            )
            right_closest = (
                np.min(orientations["right"], key=lambda x: np.abs(x - idx))
                if len(orientations["right"]) > 0
                else -1
            )

            if (left_llh.iloc[idx] > right_llh.iloc[idx]) and (
                np.abs(idx - left_closest) > np.abs(idx - right_closest)
            ):
                orientations["left"].add(idx)
                orientations["idx"].remove(idx)
            elif (left_llh.iloc[idx] > right_llh.iloc[idx]) and (
                np.abs(idx - left_closest) > np.abs(idx - right_closest)
            ):
                orientations["right"].add(idx)
                orientations["idx"].remove(idx)
            else:
                pass

        return orientations

    def _threshold_orientation(
        self, orientations, threshold, left_llh, right_llh, winsize=10
    ):
        left_llh = uniform_filter1d(left_llh, size=winsize)
        right_llh = uniform_filter1d(right_llh, size=winsize)

        left_idx = np.asarray((left_llh - right_llh) > threshold).nonzero()[0]
        right_idx = np.asarray((right_llh - left_llh) > threshold).nonzero()[0]

        orientations["left"].update(left_idx)
        orientations["right"].update(right_idx)
        orientations["idx"].difference_update(left_idx)
        orientations["idx"].difference_update(right_idx)

        return orientations

    def get_pose(self, df_coord):
        singles = self.singles

        llh, x, y = self.split_sub_df_coord(df_coord)

        if len(singles) == 0:
            self.logger.warning(
                "Empty singles body-part list in the pose configuration."
            )
            self.logger.info("Constructing pose by using all possible body-parts.")
            singles = llh.columns

        pose_dict = {}
        llh_dict = {}

        for name in singles:
            pose_dict[name + "_x"] = x[name]
            pose_dict[name + "_y"] = y[name]
            llh_dict[name] = llh[name]

        for name, components in self.defined_points.items():
            if (set(components).issubset(x.columns)) and (
                set(components).issubset(y.columns)
            ):
                def_xval = x[components].mean(axis=1)
                def_yval = y[components].mean(axis=1)
                def_llhval = llh[components].mean(axis=1)

                pose_dict[name + "_x"] = def_xval
                pose_dict[name + "_y"] = def_yval
                llh_dict[name] = def_llhval

        df_pose = pd.DataFrame.from_dict(pose_dict)
        df_llh = pd.DataFrame.from_dict(llh_dict)

        return df_pose, df_llh

    def get_oriented_pose(self, df_coord, threshold=0.5, verbose=False):
        counterparts = self.counterparts
        self.logger = logging.getLogger("main")

        def log_ckp(method, body_part, percent):
            self.logger.info(
                f"After applying {method} for {body_part}, "
                + f"orientations of {round(percent, 2)} are determined."
            )

        if len(counterparts) == 0:
            raise ValueError("Empty counterparts in the pose configuration.")

        llh, x, y = self.split_sub_df_coord(df_coord)

        pose_dict = {}
        llh_dict = {}

        for name, [left_part, right_part] in counterparts.items():
            connected_left = misc.in_nested_list(self.connected_parts, left_part)
            connected_right = misc.in_nested_list(self.connected_parts, right_part)

            left_llh = (
                llh[connected_left].mean(axis=1)
                if len(connected_left) > 0
                else llh[left_part]
            )
            right_llh = (
                llh[connected_right].mean(axis=1)
                if len(connected_right) > 0
                else llh[right_part]
            )

            # Determine orientation for each counterpart, based on likelihood values.
            orientations = {
                "left": set([]),
                "right": set([]),
                "idx": set(range(0, left_llh.shape[0])),
            }

            # (1) Applying hard threshold.
            orientations = self._threshold_orientation(
                orientations, threshold, left_llh, right_llh, winsize=6
            )
            if verbose:
                log_ckp(
                    "hard threshold comparison",
                    name,
                    1 - (len(orientations["idx"]) / left_llh.shape[0]),
                )

            # (2) In order to determine retained frames.
            orientations = self._window_count_orientation(
                orientations, left_llh, right_llh, winsize=15
            )
            if verbose:
                log_ckp(
                    "window count based comparison",
                    name,
                    1 - (len(orientations["idx"]) / left_llh.shape[0]),
                )

            # (3) If still could not be determined, compare left-right counterparts.
            orientations = self._finalize_orientation(
                orientations, left_llh, right_llh, winsize=10
            )

            if len(orientations["idx"]) > 0:
                raise AssertionError(
                    "Orientation could not be determined for some frames."
                )

            # Take coordinate values based on orientations.
            x_oriented = np.ndarray((x.shape[0]))
            y_oriented = np.ndarray((y.shape[0]))
            llh_oriented = np.ndarray((y.shape[0]))

            x_oriented[list(orientations["left"])] = x[left_part].iloc[
                list(orientations["left"])
            ]
            y_oriented[list(orientations["left"])] = y[left_part].iloc[
                list(orientations["left"])
            ]
            llh_oriented[list(orientations["left"])] = llh[left_part].iloc[
                list(orientations["left"])
            ]

            x_oriented[list(orientations["right"])] = x[right_part].iloc[
                list(orientations["right"])
            ]
            y_oriented[list(orientations["right"])] = y[right_part].iloc[
                list(orientations["right"])
            ]
            llh_oriented[list(orientations["right"])] = llh[right_part].iloc[
                list(orientations["right"])
            ]

            pose_dict[name + "_x"] = x_oriented
            pose_dict[name + "_y"] = y_oriented
            llh_dict[name] = llh_oriented

        for name, components in self.defined_points.items():
            def_xval = []
            def_yval = []
            def_llhval = []

            for item in components:
                if (item in x.columns) and (item in y.columns):
                    def_xval.append(x[item].to_numpy())
                    def_yval.append(y[item].to_numpy())
                    def_llhval.append(llh[item].to_numpy())
                if (item + "_x" in pose_dict.keys()) and (
                    item + "_y" in pose_dict.keys()
                ):
                    def_xval.append(pose_dict[item + "_x"])
                    def_yval.append(pose_dict[item + "_y"])
                    def_llhval.append(llh_dict[item])

            pose_dict[name + "_x"] = np.mean(np.array(def_xval), axis=0)
            pose_dict[name + "_y"] = np.mean(np.array(def_yval), axis=0)
            llh_dict[name] = np.mean(np.array(def_llhval), axis=0)

        df_pose, df_llh = self.get_pose(df_coord)

        df_pose_oriented = pd.DataFrame.from_dict(pose_dict)
        df_llh_oriented = pd.DataFrame.from_dict(llh_dict)

        return (
            pd.concat([df_pose, df_pose_oriented], axis=1),
            pd.concat([df_llh, df_llh_oriented], axis=1),
        )



import itertools
import logging
import math
from copy import deepcopy

import numpy as np
import pandas as pd
from scipy.ndimage.filters import uniform_filter1d

import basty.utils.misc as misc

np.seterr(all="ignore")


class SpatioTemporal:
    def __init__(self, fps, stft_cfg={}):
        self.stft_cfg = deepcopy(stft_cfg)
        self.logger = logging.getLogger("main")

        assert fps > 0
        self.get_delta = lambda x, scale: self.calc_delta(x, scale, fps)
        self.get_moving_mean = lambda x, winsize: self.calc_moving_mean(x, winsize, fps)
        self.get_moving_std = lambda x, winsize: self.calc_moving_std(x, winsize, fps)

        delta_scales_ = [100, 300, 500]
        window_sizes_ = [300, 500]

        if "delta_scales" not in stft_cfg.keys():
            self.logger.info(
                "Scale valuess can not be found in configuration for delta features."
                + f"Default values are {str(delta_scales_)[1:-1]}."
            )

        if "window_sizes" not in stft_cfg.keys():
            self.logger.info(
                "Window sizes can not be found in configuration for window features."
                + f"Default values are {str(window_sizes_)[1:-1]}."
            )

        self.stft_cfg["delta_scales"] = stft_cfg.get("delta_scales", delta_scales_)
        self.stft_cfg["window_sizes"] = stft_cfg.get("window_sizes", window_sizes_)

        self.stft_set = ["pose", "distance", "angle"]
        for ft_set in self.stft_set:
            ft_set_dt = ft_set + "_delta"
            self.stft_cfg[ft_set] = stft_cfg.get(ft_set, [])
            self.stft_cfg[ft_set_dt] = stft_cfg.get(ft_set_dt, [])

        self.angle_between = self.angle_between_atan

    @staticmethod
    def angle_between_arccos(v1, v2):
        """
        Returns the abs(angle) in radians between vectors 'v1' and 'v2'.
        angle_between((1, 0, 0), (0, 1, 0)) --> 1.5707963267948966
        angle_between((1, 0, 0), (1, 0, 0)) --> 0.0
        angle_between((1, 0, 0), (-1, 0, 0)) --> 3.141592653589793
        """
        assert isinstance(v1, np.ndarray) and isinstance(v2, np.ndarray)

        v1_u = v1 / np.linalg.norm(v1)
        v2_u = v2 / np.linalg.norm(v2)

        return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))

    @staticmethod
    def angle_between_atan(v1, v2):
        """
        Returns the abs(angle) in radians between vectors 'v1' and 'v2'.
        """
        assert isinstance(v1, np.ndarray) and isinstance(v2, np.ndarray)

        angle = np.math.atan2(np.linalg.det([v1, v2]), np.dot(v1, v2))

        return np.abs(angle)

    def get_group_value(self, stft_group, opt):
        if opt == "avg":
            group_value = np.nanmean(stft_group, axis=1)
        elif opt == "min":
            group_value = np.nanamin(stft_group, axis=1)
        elif opt == "max":
            group_value = np.nanmax(stft_group, axis=1)
        else:
            raise ValueError(f"Unkown option {opt} is given for feature group.")
        return group_value

    @staticmethod
    def calc_delta(x, scale, fps):
        # In terms of millisecond.
        delta_values = []
        scale_frame = math.ceil(fps * (1000 / scale))

        y = uniform_filter1d(x, size=scale_frame, axis=0)
        delta_y = np.abs(np.gradient(y, 1 / fps * 1000, axis=0, edge_order=2))
        delta_values.append(delta_y)

        return delta_values

    @staticmethod
    def calc_moving_mean(x, winsize, fps):
        mean_values = []
        w_frame = math.ceil(fps * (winsize / 1000))

        mean_values.append(x.rolling(w_frame, min_periods=1, center=True).mean())

        return mean_values

    @staticmethod
    def calc_moving_std(x, winsize, fps):
        std_values = []
        w_frame = math.ceil(fps * (winsize / 1000))

        std_values.append(x.rolling(w_frame, min_periods=1, center=True).std())

        return std_values

    def extract(self, ft_set, df_pose, ft_cfg_set):
        extraction_functions = {
            "pose": self._extract_pose,
            "angle": self._extract_angle,
            "distance": self._extract_distance,
        }
        val = extraction_functions[ft_set](df_pose, ft_cfg_set)
        return val

    def get_column_names(self, ft_set):
        stft_cfg = self.stft_cfg
        name_col = []

        def get_stft_name(defn):
            if isinstance(defn, dict):
                name = (
                    list(defn.keys())[0]
                    + "("
                    + ",".join(["-".join(item) for item in list(defn.values())[0]])
                    + ")"
                )
            elif isinstance(defn, list):
                name = "-".join(defn)
            else:
                raise ValueError(
                    f"Given feature definition {defn} has incorrect formatting."
                )
            return name

        if not stft_cfg.get(ft_set, False):
            raise ValueError(f"Unkown value {ft_set} is given for feature set.")

        if "pose" in ft_set:
            ft_names = list(
                itertools.chain.from_iterable(
                    ([item + "_x"], [item + "_y"]) for item in stft_cfg[ft_set]
                )
            )
        else:
            ft_names = stft_cfg[ft_set]

        if "delta" not in ft_set:
            name_col = [ft_set + "." + get_stft_name(item) for item in ft_names]

        else:
            scales = stft_cfg["delta_scales"]
            name_col = misc.flatten(
                [
                    [
                        ft_set + "." + get_stft_name(item) + ".s" + str(t)
                        for item in ft_names
                    ]
                    for t in scales
                ]
            )

        return name_col

    @staticmethod
    def _get_coord(df_pose, name, axis):
        # Axis name x or y.
        name_c = name + "_" + axis

        if name_c in df_pose.columns:
            coord = df_pose[name_c]
        elif name == "origin":
            coord = np.zeros(df_pose.shape[0])
        else:
            raise ValueError(f"No coordinate values can be found for {name}.")

        return coord

    def _extract_pose(self, df_pose, body_parts):
        xy_pose_values = np.ndarray((df_pose.shape[0], len(body_parts) * 2))

        if not isinstance(body_parts, list):
            raise ValueError(
                f"Given argument has type {type(body_parts)}."
                + "Pose features should be defined by a list of body-parts."
            )

        for i, bp in enumerate(body_parts):
            if not isinstance(bp, str):
                raise ValueError(
                    f"Given feature definition contains {bp}, which is not a body-part."
                )
            xy_pose_values[:, i * 2] = self.__class__._get_coord(df_pose, bp, "x")
            xy_pose_values[:, i * 2 + 1] = self.__class__._get_coord(df_pose, bp, "y")

        return xy_pose_values

    def _extract_angle(self, df_pose, triplets):
        angle_values = np.ndarray((df_pose.shape[0], len(triplets)))

        def f_angle(x):
            return self.angle_between(x[:2] - x[2:4], x[4:] - x[2:4])

        def angle_along_axis(xy_values, angle_values):
            for j in range(xy_values.shape[0]):
                v1 = xy_values[j, :2] - xy_values[j, 2:4]
                v2 = xy_values[j, 4:] - xy_values[j, 2:4]
                angle_values[j, i] = self.angle_between(v1, v2)
            return angle_values

        for i, triplet in enumerate(triplets):
            if isinstance(triplet, dict):
                opt = list(triplet.keys())[0]
                group = list(triplet.values())[0]
                if len(group) > 0 and opt in ["avg", "min", "max"]:
                    angle_group = self._extract_angle(df_pose, group)
                else:
                    raise ValueError(f"Given feature definition {triplet} is unknown.")
                angle_values[:, i] = self.get_group_value(angle_group, opt)
            else:
                xy_values, _ = self._extract_pose(df_pose, triplet)
                # angle_values[:, i] = np.apply_along_axis(f_angle, 1, xy_values)
                # This is somehow faster.
                angle_values[:, i] = angle_along_axis(xy_values, angle_values)

        return angle_values

    def _extract_distance(self, df_pose, pairs):
        distance_values = np.ndarray((df_pose.shape[0], len(pairs)))

        for i, pair in enumerate(pairs):
            if isinstance(pair, dict):
                opt = list(pair.keys())[0]
                group = list(pair.values())[0]
                if len(group) > 0 and opt in ["avg", "min", "max"]:
                    distance_group = self._extract_distance(df_pose, group)
                else:
                    raise ValueError(f"Given feature definition {pair} is unkwon.")
                distance_values[:, i] = self.get_group_value(distance_group, opt)
            else:
                xy_values = self._extract_pose(df_pose, pair)
                diff_xy = xy_values[:, 2:4] - xy_values[:, :2]
                distance_values[:, i] = np.sqrt(diff_xy[:, 0] ** 2 + diff_xy[:, 1] ** 2)

        return distance_values

    def _extract_moving_stat(self, df_stft, stft_names_dict, stat, winsizes):
        if stat == "mean":
            get_moving_stat = self.get_moving_mean
        elif stat == "std":
            get_moving_stat = self.get_moving_std
        else:
            raise ValueError(f"Unkown value {stat} is given for moving statistics.")

        name_col = df_stft.columns

        mv_stat = pd.concat(
            itertools.chain(*map(lambda w: get_moving_stat(df_stft, w), winsizes)),
            axis=1,
        )
        df_stat = pd.DataFrame(data=mv_stat)

        stat_columns = misc.flatten(
            [
                [
                    stat + "." + stft_names_dict[name] + ".w" + str(w)
                    for name in name_col
                ]
                for w in winsizes
            ]
        )
        name_dict = {i: stat_columns[i] for i in range(len(stat_columns))}
        df_stat.columns = list(name_dict.keys())

        return df_stat, name_dict

    def extract_snap_stft(self, df_pose):
        stft_cfg = self.stft_cfg

        df_snap_list = []

        for ft_set in self.stft_set:
            if stft_cfg.get(ft_set, False):
                temp_df = pd.DataFrame(self.extract(ft_set, df_pose, stft_cfg[ft_set]))
                temp_df.columns = self.get_column_names(ft_set)
                df_snap_list.append(temp_df)

        if len(df_snap_list) <= 0:
            raise ValueError(
                "At least one snap feature must given in the feature configuration."
            )

        df_snap = pd.concat(df_snap_list, axis=1)

        name_col = df_snap.columns
        name_dict = {i: name_col[i] for i in range(len(name_col))}
        df_snap.columns = list(name_dict.keys())
        self.ftname_to_snapft = name_dict

        return df_snap, name_dict

    def extract_delta_stft(self, df_pose):
        stft_cfg = self.stft_cfg
        delta_scales = stft_cfg["delta_scales"]

        df_delta_list = []

        for ft_set in self.stft_set:
            ft_set_dt = ft_set + "_delta"

            if stft_cfg.get(ft_set_dt, False):
                temp_snap = self.extract(ft_set, df_pose, stft_cfg[ft_set_dt])
                temp_delta = itertools.chain(
                    *map(
                        lambda s: self.get_delta(temp_snap, s),
                        delta_scales,
                    )
                )

                temp_df = pd.DataFrame(
                    np.concatenate(
                        tuple(temp_delta),
                        axis=1,
                    ),
                    columns=self.get_column_names(ft_set_dt),
                )
                df_delta_list.append(temp_df)

        if len(df_delta_list) <= 0:
            raise ValueError(
                "At least one delta feature must given in the feature configuration."
            )

        df_delta = pd.concat(df_delta_list, axis=1)

        name_col = df_delta.columns
        name_dict = {i: name_col[i] for i in range(len(name_col))}
        df_delta.columns = list(name_dict.keys())
        self.ftname_to_deltaft = name_dict

        return df_delta, name_dict

    def extract_window_snap_stft(self, df_stft, opt):
        window_sizes = self.stft_cfg["window_sizes"]
        if opt == "mean":
            df_window, name_dict = self._extract_moving_stat(
                df_stft, self.ftname_to_snapft, "mean", window_sizes
            )
            self.ftname_to_snapft_mean = name_dict
        elif opt == "std":
            df_window, name_dict = self._extract_moving_stat(
                df_stft, self.ftname_to_snapft, "std", window_sizes
            )
            self.ftname_to_snapft_std = name_dict
        return df_window, name_dict

    def extract_window_delta_stft(self, df_stft, opt):
        window_sizes = self.stft_cfg["window_sizes"]
        if opt == "mean":
            df_window, name_dict = self._extract_moving_stat(
                df_stft, self.ftname_to_deltaft, "mean", window_sizes
            )
            self.ftname_to_deltaft_mean = name_dict
        elif opt == "std":
            df_window, name_dict = self._extract_moving_stat(
                df_stft, self.ftname_to_deltaft, "std", window_sizes
            )
            self.ftname_to_deltaft_std = name_dict
        return df_window, name_dict



import logging
import sys

import joblib
import numpy as np
import pandas as pd
import tqdm

import basty.utils.io as io


class ProgressBarHandler(logging.Handler):
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)

    def emit(self, record):
        try:
            msg = self.format(record)
            tqdm.tqdm.write(msg)
            self.flush()
        except Exception:
            self.handleError(record)


class ParameterHandler:
    def __init__(self):
        pass

    def init_annotation_kwargs(self, **kwargs):
        self.annotation_priority = kwargs.pop("annotation_priority", [])
        self.inactive_annotation = kwargs.pop("inactive_annotation", "Idle&Other")
        self.noise_annotation = kwargs.pop("noise_annotation", "Noise")
        self.arouse_annotation = kwargs.pop("arouse_annotation", "Moving")

    def init_preprocessing_kwargs(self, **kwargs):
        self.compute_oriented_pose = kwargs.pop("compute_oriented_pose", True)
        self.compute_egocentric_frames = kwargs.pop("compute_egocentric_frames", False)
        self.save_likelihood = kwargs.pop("save_likelihood", False)

        self.local_outlier_threshold = kwargs.pop("local_outlier_threshold", 15)
        assert self.local_outlier_threshold > 0
        self.local_outlier_winsize = kwargs.pop("local_outlier_winsize", 15)

        self.decreasing_llh_winsize = kwargs.pop("decreasing_llh_winsize", 0)
        self.decreasing_llh_lower_z = kwargs.pop("decreasing_llh_lower_z", -3)

        self.low_llh_threshold = kwargs.pop("low_llh_threshold", 0.05)
        assert self.low_llh_threshold < 1

        self.median_filter_winsize = kwargs.pop("median_filter_winsize", 3)
        self.boxcar_filter_winsize = kwargs.pop("boxcar_filter_winsize", 0)

        self.jump_quantile = kwargs.pop("jump_quantile", 0)

        self.interpolation_method = kwargs.pop("interpolation_method", "linear")
        self.interpolation_kwargs = kwargs.pop("interpolation_kwargs", {})
        self.kalman_filter_kwargs = kwargs.pop("kalman_filter_kwargs", {})

    def init_behavioral_reprs_kwargs(self, **kwargs):
        self.use_cartesian_blent = kwargs.pop("use_cartesian_blent", False)
        self.norm = kwargs.pop("norm", "l1")
        assert self.norm in ["l1", "l2", "max"]

    @staticmethod
    def gradient_boosting_default_kwargs(**kwargs):
        gradient_boosting_kwargs = {}
        gradient_boosting_kwargs["n_estimators"] = kwargs.pop("n_estimators", 10)
        gradient_boosting_kwargs["max_depth"] = kwargs.pop("max_depth", 5)
        gradient_boosting_kwargs["min_samples_leaf"] = kwargs.pop(
            "min_samples_leaf", 10**3
        )
        gradient_boosting_kwargs["max_features"] = kwargs.pop("max_features", "sqrt")
        gradient_boosting_kwargs["criterion"] = kwargs.pop("criterion", "friedman_mse")
        return gradient_boosting_kwargs

    @staticmethod
    def decision_tree_default_kwargs(**kwargs):
        decision_tree_kwargs = {}
        decision_tree_kwargs["n_estimators"] = kwargs.pop("n_estimators", 5)
        decision_tree_kwargs["max_depth"] = kwargs.pop("max_depth", 5)
        decision_tree_kwargs["min_samples_leaf"] = kwargs.pop(
            "min_samples_leaf", 10**2
        )
        decision_tree_kwargs["max_features"] = kwargs.pop("max_features", "sqrt")
        decision_tree_kwargs["criterion"] = kwargs.pop("criterion", "gini")
        decision_tree_kwargs["class_weight"] = kwargs.pop(
            "class_weight", "balanced_subsample"
        )
        return decision_tree_kwargs

    @staticmethod
    def nearest_neighbors_default_kwargs(**kwargs):
        nearest_neighbors_kwargs = {}
        nearest_neighbors_kwargs["n_neighbors"] = kwargs.pop("n_neighbors", 50)
        nearest_neighbors_kwargs["weights"] = kwargs.pop("weights", "distance")
        nearest_neighbors_kwargs["algorithm"] = kwargs.pop("algorithm", "auto")
        nearest_neighbors_kwargs["n_jobs"] = kwargs.pop("n_jobs", -1)
        return nearest_neighbors_kwargs

    def init_dormant_epochs_kwargs(self, fps, use_supervised_learning, **kwargs):
        self.datums = kwargs.pop("datums", [])
        self.datums_winsize = kwargs.pop("datums_winsize", fps * 1)
        self.scale = kwargs.pop("scale", False)
        self.log_scale = kwargs.pop("log_scale", False)

        self.min_dormant = kwargs.pop("min_dormant", 300 * fps)

        if not use_supervised_learning:
            self.num_gmm_comp = kwargs.pop("num_gmm_comp", 2)
            self.threshold_key = kwargs.pop("threshold_key", "local_max")
            # Indices start from 0.
            self.threshold_idx = kwargs.pop("threshold_idx", 1)

            self.epoch_winsize = kwargs.pop("epoch_winsize", 180 * fps)
            self.tol_duration = kwargs.pop("tol_duration", 90 * fps)
            self.tol_percent = kwargs.pop("tol_percent", 0.4)
        else:
            self.post_processing_winsize = kwargs.pop(
                "post_processing_winsize", fps * 2
            )
            self.post_processing_wintype = kwargs.pop(
                "post_processing_wintype", "boxcar"
            )
            self.classifier_kwargs = self.__class__.decision_tree_default_kwargs(
                **kwargs
            )
            # self.classifier_kwargs = self.__class__.nearest_neighbors_default_kwargs(
            #     **kwargs
            # )
            self.label_conversion_dict = kwargs.pop("label_conversion_dict", {})

    def init_active_bouts_kwargs(self, fps, use_supervised_learning, **kwargs):
        self.datums_list = kwargs.pop("datums_list", [[]])
        self.datums_winsize = kwargs.pop("datums_winsize", fps // 10)
        self.coefs_summary_method = kwargs.pop("coefs_summary_method", "max")
        self.scale = kwargs.pop("scale", True)
        self.log_scale = kwargs.pop("log_scale", True)

        self.post_processing_winsize = kwargs.pop("post_processing_winsize", fps * 1)
        self.post_processing_wintype = kwargs.pop("post_processing_wintype", "boxcar")

        if not use_supervised_learning:
            self.num_gmm_comp = kwargs.pop("num_gmm_comp", 3)
            self.threshold_key = kwargs.pop("threshold_key", "local_max")
            # Indices start from 0.
            self.threshold_idx = kwargs.pop("threshold_idx", 1)
        else:
            self.classifier_kwargs = self.__class__.decision_tree_default_kwargs(
                **kwargs
            )
            # self.classifier_kwargs = self.__class__.nearest_neighbors_default_kwargs(
            #     **kwargs
            # )
            self.label_conversion_dict = kwargs.pop("label_conversion_dict", {})

    def init_behavior_embeddings_kwargs(self, **kwargs):
        self.UMAP_kwargs = {}
        self.UMAP_kwargs["n_neighbors"] = kwargs.pop("embedding_n_neighbors", 90)
        self.UMAP_kwargs["min_dist"] = kwargs.pop("embedding_min_dist", 0.0)
        self.UMAP_kwargs["spread"] = kwargs.pop("embedding_spread", 1.0)
        self.UMAP_kwargs["n_components"] = kwargs.pop("embedding_n_components", 2)
        self.UMAP_kwargs["metric"] = kwargs.pop("embedding_metric", "hellinger")
        self.UMAP_kwargs["low_memory"] = kwargs.pop("embedding_low_memory", True)
        self.use_annotations_to_mask = kwargs.pop(
            "use_annotations_to_mask", (False, False)
        )
        assert (
            isinstance(self.use_annotations_to_mask, tuple)
            and len(self.use_annotations_to_mask) == 2
        )

    def init_behavior_clustering_kwargs(self, **kwargs):
        self.HDBSCAN_kwargs = {}
        self.HDBSCAN_kwargs["prediction_data"] = kwargs.pop("prediction_data", True)
        self.HDBSCAN_kwargs["approx_min_span_tree"] = kwargs.pop(
            "approx_min_span_tree", True
        )
        self.HDBSCAN_kwargs["cluster_selection_method"] = kwargs.pop(
            "cluster_selection_method", "eom"
        )
        self.HDBSCAN_kwargs["cluster_selection_epsilon"] = kwargs.pop(
            "cluster_selection_epsilon", 0.0
        )
        self.HDBSCAN_kwargs["min_cluster_size"] = kwargs.pop("min_cluster_size", 500)
        self.HDBSCAN_kwargs["min_samples"] = kwargs.pop("min_cluster_samples", 5)

    def init_behavior_mapping_postprocessing_kwargs(self, **kwargs):
        pass

    def init_behavior_correspondence_kwargs(self, **kwargs):
        pass


class Logger:
    def __init__(self, stdout=False, filepath=None, logformat=None):
        self.stdout = stdout
        self.filepath = filepath
        self.logformat = logformat
        self.handlers = []
        self.handlers.append(ProgressBarHandler())

        if self.logformat is None:
            self.logformat = "[%(asctime)s] %(levelname)s - %(message)s"
            self.datefmt = "%Y-%m-%d %H:%M:%S"
        if self.stdout:
            stdout_handler = logging.StreamHandler(sys.stdout)
            self.handlers.append(stdout_handler)
        if self.filepath is not None:
            io.ensure_file_dir(self.filepath)
            file_handler = logging.FileHandler(filename=filepath)
            self.handlers.append(file_handler)

        logging.basicConfig(
            format=self.logformat,
            handlers=self.handlers,
            level=logging.INFO,
            datefmt=self.datefmt,
        )
        self.logger = logging.getLogger("main")

    def direct_error(self, err):
        self.logger.error(err)

    def direct_warning(self, warn):
        self.logger.warning(warn)

    def direct_info(self, info):
        self.logger.info(info)

    @staticmethod
    def _get_path_str(path, depth):
        return "/".join(path.parts[-depth:-1])

    # Log Messages:
    def file_already_exists(self, name, path, depth=2):
        output_path = self._get_path_str(path, depth)
        self.logger.warning(f"FAE - {name} already exists in {output_path}, skipped.")

    def file_saved(self, name, path, depth=2):
        output_path = self._get_path_str(path, depth)
        self.logger.info(f"FS - {name} has been saved to {output_path}.")

    def file_missing(self, name, path, depth=2):
        output_path = self._get_path_str(path, depth)
        self.logger.error(f"FM - {name} can not be found in {output_path}.")

    def not_initialized(self, name):
        self.logger.error(f"NI - {name} is not initialized yet.")

    def invalid_value_warning(self, name, count):
        self.logger.warning(f"IVW - {name} contains {str(count)} inf/nan values.")

    def type_error(self, name, type_name):
        self.logger.error(f"TE - type of '{name}' must be {type_name}.")

    def missing_cfg_key_error(self, name):
        self.logger.error(
            f"MCKE - key {name} can not be found in the configuration file."
        )

    def missing_cfg_key_warning(self, name):
        self.logger.warning(
            f"MCKW - key {name} can not be found in the configuration file, "
            f"default value will be used."
        )


class LoadingHelper:
    def __init__(self, logger):
        self.logger = logger

    def _load_numpy_array(self, path, name):
        npy_path = path / name
        try:
            npy = np.load(npy_path)
        except FileNotFoundError as err:
            self.logger.direct_error(err)
            self.logger.file_missing(f"array {name}", npy_path)
            raise
        return npy

    def _load_joblib_object(self, path, name):
        obj_path = path / name
        try:
            obj = joblib.load(obj_path)
        except FileNotFoundError as err:
            self.logger.direct_error(err)
            self.logger.file_missing(f"object {name}", obj_path)
            raise
        return obj

    def _load_pandas_dataframe(self, path, name):
        df_path = path / name
        try:
            df = pd.read_pickle(df_path)
        except FileNotFoundError as err:
            self.logger.direct_error(err)
            self.logger.file_missing(f"dataframe {name}", df_path)
            raise
        return df

    def _load_yaml_dictionary(self, expt_path, name):
        dict_path = expt_path / name
        try:
            dict_ = io.read_yaml(dict_path)
        except FileNotFoundError as err:
            self.logger.direct_error(err)
            self.logger.file_missing(f"dictionary {name}", dict_path)
            raise
        return dict_


class SavingHelper:
    def __init__(self, logger):
        self.logger = logger

    def _save_numpy_array(self, npy, path, name, msg="", depth=2):
        npy_path = path / name
        np.save(npy_path, npy)
        log_msg = f"{name} {msg}" if len(msg) > 0 else name
        self.logger.file_saved(f"array {log_msg}", npy_path, depth=depth)

    def _save_joblib_object(self, obj, path, name, msg="", depth=2):
        obj_path = path / name
        joblib.dump(obj, obj_path)
        log_msg = f"{name} {msg}" if len(msg) > 0 else name
        self.logger.file_saved(f"object {log_msg}", obj_path, depth=depth)

    def _save_pandas_dataframe(self, df, path, name, msg="", depth=2):
        df_path = path / name
        df.to_pickle(df_path)
        log_msg = f"{name} {msg}" if len(msg) > 0 else name
        self.logger.file_saved(f"dataframe {log_msg}", df_path, depth=depth)

    def _save_yaml_dictionary(self, dict_obj, expt_path, name, msg="", depth=2):
        dict_path = expt_path / name
        io.dump_yaml(dict_obj, dict_path)
        log_msg = f"{name} {msg}" if len(msg) > 0 else name
        self.logger.file_saved(f"yaml file {log_msg}", dict_path, depth=depth)



import re
from collections import defaultdict

import numpy as np
import pandas as pd
from sklearn.preprocessing import Normalizer
from tqdm import tqdm

import basty.utils.misc as misc
import basty.utils.preprocessing as prep
from basty.feature_extraction.body_pose import BodyPose
from basty.feature_extraction.spatiotemporal_features import SpatioTemporal
from basty.feature_extraction.wavelet_transformation import WaveletTransformation
from basty.project.experiment_processing import Project


class FeatureExtraction(Project):
    def __init__(self, main_cfg_path, **kwargs):
        Project.__init__(self, main_cfg_path, **kwargs)
        self.init_preprocessing_kwargs(**kwargs)
        self.init_behavioral_reprs_kwargs(**kwargs)

    @misc.timeit
    def compute_pose_values(self):
        bp = BodyPose(self.pose_cfg)

        pbar = tqdm(self.data_path_dict.items())
        for name, data_path in pbar:
            pbar.set_description(f"Computing pose values for experiment {name}")

            expt_path = self.expt_path_dict[name]
            expt_record = self._load_joblib_object(expt_path, "expt_record.z")

            expt_idx_key = []
            df_pose_list = []
            df_llh_list = []

            expt_pose_path = expt_path / "pose.pkl"
            if expt_pose_path.exists():
                self.logger.file_already_exists("pose.pkl", expt_pose_path)
            else:
                file_path_list = list(data_path.glob(name + ".csv"))
                # At least one .csv file  with appropriate naming must be found
                # in the given experiment data directory.
                assert len(file_path_list) != 0

                for file_path in tqdm(sorted(file_path_list), leave=False, desc=name):
                    expt_partID = file_path.stem.split("-")[-1]
                    expt_idx = (
                        int(expt_partID)
                        if re.compile("[0-9]" * 4).match(expt_partID)
                        else 0
                    )
                    expt_idx_key.append(int(expt_idx))

                    df_coord = pd.read_csv(file_path, low_memory=False)

                    if self.compute_oriented_pose:
                        pose_tmp, llh_tmp = bp.get_oriented_pose(df_coord)
                    else:
                        pose_tmp, llh_tmp = bp.get_pose(df_coord)

                    df_llh_list.append(llh_tmp)

                    if self.compute_egocentric_frames:
                        df_pose_list.append(bp.make_frames_egocentric(pose_tmp))
                    else:
                        df_pose_list.append(pose_tmp)

                expt_idx_key = np.array(expt_idx_key)
                assert expt_idx_key.shape[0] == 1 or (np.diff(expt_idx_key) == 1).all()

                df_llh = pd.concat(
                    df_llh_list,
                    keys=expt_idx_key,
                    names=["video-part", "frame-index"],
                )

                if self.save_likelihood:
                    self._save_pandas_dataframe(df_llh, expt_path, "llh.pkl")

                def pose_interpolate(df_pose):
                    df_pose.interpolate(
                        method=self.interpolation_method,
                        **self.interpolation_kwargs,
                        axis=0,
                        inplace=True,
                    )
                    df_pose = prep.fill_interpolate(df_pose)
                    return df_pose

                df_pose = pd.concat(
                    df_pose_list, keys=expt_idx_key, names=["video-part", "frame-index"]
                )
                df_pose.reset_index(inplace=True)
                vcols = [
                    col
                    for col in df_pose.columns
                    if col not in ["video-part", "frame-index"]
                ]

                if df_pose[vcols].isnull().values.any():
                    df_pose = pose_interpolate(df_pose[vcols])

                if self.local_outlier_winsize > 0:
                    df_pose[vcols] = prep.remove_local_outliers(
                        df_pose[vcols],
                        winsize=self.local_outlier_winsize,
                        threshold=self.local_outlier_threshold,
                    )

                if self.decreasing_llh_winsize > 0:
                    df_pose[vcols], _ = prep.remove_decrasing_llh_frames(
                        df_pose[vcols],
                        df_llh,
                        winsize=self.decreasing_llh_winsize,
                        lower_z=self.decreasing_llh_lower_z,
                    )

                df_pose[vcols], _ = prep.remove_low_llh_frames(
                    df_pose[vcols], df_llh, threshold=self.low_llh_threshold
                )
                df_pose.replace([np.inf, -np.inf], np.nan, inplace=True)

                df_pose[vcols] = pose_interpolate(df_pose[vcols])

                if self.median_filter_winsize > 0:
                    df_pose[vcols] = prep.median_filter(
                        df_pose[vcols], self.median_filter_winsize
                    )

                # Interpolate jumps based on the gradient quantiles.
                if self.jump_quantile > 0:
                    df_pose[vcols] = prep.remove_jumps(
                        df_pose[vcols], q=self.jump_quantile
                    )
                    df_pose[vcols] = pose_interpolate(df_pose[vcols])

                if self.boxcar_filter_winsize > 0:
                    df_pose[vcols] = prep.boxcar_center_filter(
                        df_pose[vcols], self.boxcar_filter_winsize
                    )

                if self.kalman_filter_kwargs:
                    df_pose = prep.rts_smoother_one_dimensional(
                        df_pose, self.kalman_filter_kwargs, dim_x=2, dim_z=1
                    )

                # Fix and name multi-index.
                df_pose.set_index(["video-part", "frame-index"], inplace=True)
                for i, lvl in enumerate(df_pose.index.levels):
                    df_pose.index = df_pose.index.set_levels(lvl.astype(int), level=i)

                # Check and count NaN and Inf values.
                count = np.isinf(df_pose).values.sum()
                count += np.isnan(df_pose).values.sum()
                if count > 0:
                    self.logger.invalid_value_warning("pose values", count)

                self._save_pandas_dataframe(df_pose, expt_path, "pose.pkl")

                group_indices = df_pose.groupby(level=0, as_index=True).indices
                part_to_frame_count = {
                    key: max(val) + 1 for key, val in group_indices.items()
                }

                expt_record.part_to_frame_count = part_to_frame_count
                self._save_joblib_object(
                    expt_record,
                    expt_path,
                    "expt_record.z",
                    "with the frame count information",
                )

    @misc.timeit
    def compute_spatiotemporal_features(self, delta=True, snap=True):
        stft_extractor = SpatioTemporal(self.fps, self.feature_cfg)

        pbar = tqdm(self.expt_path_dict.items())
        for name, expt_path in pbar:
            pbar.set_description(f"Computing spatiotempral features of {name}")

            df_pose = self._load_pandas_dataframe(expt_path, "pose.pkl")

            if delta:
                delta_path = expt_path / "delta_stft.pkl"
                if delta_path.exists():
                    self.logger.file_already_exists(
                        "delta spatio-temporal features", delta_path
                    )
                else:
                    df_delta, ftname_to_deltaft = stft_extractor.extract_delta_stft(
                        df_pose
                    )
                    self._save_yaml_dictionary(
                        ftname_to_deltaft, expt_path, "ftname_to_deltaft.yaml"
                    )
                    self._save_pandas_dataframe(df_delta, expt_path, "delta_stft.pkl")

            if snap:
                snap_path = expt_path / "snap_stft.pkl"
                if snap_path.exists():
                    self.logger.file_already_exists(
                        "snap spatio-temporal features", snap_path
                    )
                else:
                    df_snap, ftname_to_snapft = stft_extractor.extract_snap_stft(
                        df_pose
                    )
                    self._save_yaml_dictionary(
                        ftname_to_snapft, expt_path, "ftname_to_snapft.yaml"
                    )
                    self._save_pandas_dataframe(df_snap, expt_path, "snap_stft.pkl")

    @misc.timeit
    def compute_postural_dynamics(self):
        wt_detached_cfg = self.temporal_cfg.get("wt_detached", {})
        wt_cfg = self.temporal_cfg.get("wt", {})

        pbar = tqdm(self.expt_path_dict.items())
        for name, expt_path in pbar:
            pbar.set_description(f"Computing postural dynamics of {name}")

            ws_path = expt_path / "wsft.npy"
            if ws_path.exists():
                self.logger.file_already_exists("wavelet spectrum values", ws_path)
            else:
                snap_stft = self._load_pandas_dataframe(expt_path, "snap_stft.pkl")
                snap_stft = snap_stft.to_numpy()
                ftname_to_snapft = self._load_yaml_dictionary(
                    expt_path, "ftname_to_snapft.yaml"
                )
                wsft_list = []

                for key in ftname_to_snapft.keys():
                    idx = key
                    snap_wv_cfg = wt_detached_cfg.get(key, wt_cfg)

                    if not snap_wv_cfg:
                        self.missing_cfg_key_error(
                            self, "wavelet transformation configuration"
                        )

                    wvt = WaveletTransformation(self.fps, snap_wv_cfg)
                    ftws_r = wvt.cwtransform(snap_stft[:, [idx]])
                    wsft_list.append(wvt.normalize_channels(ftws_r))

                wsft = np.concatenate(wsft_list, axis=-1)
                self._save_numpy_array(wsft, expt_path, "wsft.npy")

    @staticmethod
    def compute_cartesian_blent(wsft, ftname_to_snapft):
        component_dict = defaultdict(list)
        new_wsft_list = []

        for i, val in enumerate(ftname_to_snapft.items()):
            name_split = val.split(".")
            if name_split[0] == "pose":
                bp_name = name_split[-1].split("_")[0]
                component_dict[bp_name].append(i)
            else:
                new_wsft_list.append(wsft[:, :, i])

        for _, val in component_dict.items():
            new_wsft_list.append(np.sum(wsft[:, :, val], axis=2) / len(val))

        return np.stack(new_wsft_list, axis=-1)

    @misc.timeit
    def compute_behavioral_representations(self):
        pbar = tqdm(self.expt_path_dict.items())
        for name, expt_path in pbar:
            pbar.set_description(f"Computing behavioral representations of {name}")

            expt_path = self.expt_path_dict[name]

            wsft = self._load_numpy_array(expt_path, "wsft.npy")
            if self.use_cartesian_blent:
                ftname_to_snapft = self._load_yaml_dictionary(
                    expt_path, "ftname_to_snapft.yaml"
                )
                wsft = self.__class__.compute_cartesian_blent(wsft, ftname_to_snapft)

            flat_dim = wsft.shape[1] * wsft.shape[2]
            ftws_flat = np.reshape(wsft, (wsft.shape[0], flat_dim))

            l1_normalizer = Normalizer(norm=self.norm)
            X_expt = l1_normalizer.transform(ftws_flat)

            self._save_numpy_array(X_expt, expt_path, "behavioral_reprs.npy")



import itertools
import logging
import os
from pathlib import Path
import subprocess
import textwrap
import time
from collections import Counter
import yaml
import basty.feature_extraction.body_pose as body_pose
import re

import numpy as np
import pandas as pd


def list_cartesian_product(*args):
    assert all([isinstance(list_arg, list) for list_arg in args])
    return itertools.product(*args)


def dict_check_and_sort(x):
    if x is not None:
        assert isinstance(x, dict)
        x = dict(sorted(x.items()))
    return x


def txt_wrap(x, lw=72):
    return textwrap.fill(x, lw)


def sort_dict(x):
    if x is not None:
        # Sorts dictionary by key.
        return dict(sorted(x.items()))
    return x


def parse_experiment_names(data_path_dict):
    """Takes in the name of the experiments and returns a dataframe containing
    sex,age and sleep deprivation info"""
    df = pd.DataFrame(data_path_dict.items(), columns=['ExptNames', 'Path'])
    df['SD'] = df['ExptNames'].str.contains('_SD')
    df['Age'] = df['ExptNames'].str.extract(re.compile("(_[1-9]|10)[dD]"))
    df['Age'] = pd.to_numeric(df['Age'].str[1])
    df['Sex'] = df['ExptNames'].str.extract(r"_([FM])")
    # Assign 5 to 'NaN' age (if age is not listed, then it is 5 days old per expt protocols)
    df['Age'] = df['Age'].fillna(5)

    def check_folders(folderpath):
        txt_file = [file for file in os.listdir(folderpath) if 'CB' in file and file.endswith('.txt')]
        if txt_file:
            if 'annotated' not in txt_file[0]:
                experimenter_name = os.path.splitext(txt_file[0])[0]
            else:
                experimenter_name = 'MK'
        else:
            experimenter_name = 'MK'
        return experimenter_name

    # Finaly check if each folder contains CB.txt files which indicates the experimenter. Assume MK if it is not existent
    df['Experimenter'] = df['Path'].apply(lambda x: check_folders(x))
    return df


def split_video_by_annotation(idxpath, vidname):
    """
    Takes in a file containing list of indices and a video file, returns
    smaller videos corresponding to the list.
    """
    df = pd.read_csv(idxpath)
    bout = 0
    fps = 30
    targetfolder = os.path.join(os.path.dirname(idxpath), "SplitVids")
    os.mkdir(targetfolder)

    for start, finish, behavior in zip(df.Start, df.Finish, df.Behavior):
        idx1 = start / fps
        idx2 = finish / fps
        tempname = behavior + "_" + str(bout) + "_" + str(start) + "_" + str(finish)
        outputname = os.path.join(targetfolder, tempname + ".avi")
        bout += 1
        command = (
            f"ffmpeg -n -i {vidname} -ss {idx1} -to {idx2} " f"-c:v copy {outputname}"
        )
        subprocess.call(command, shell=True)


def organize_file_structure(path):
    """
    Modifies the filenames of .csv files to match it with videos and returns a dictionary to be incorporated in the main_cfg.yaml
    """
    new_csv_names = []
    folder_dict = {}

    # Iterate over folders
    for dirs in os.listdir(path):
        avi_names = ([f for f in os.listdir(os.path.join(path, dirs)) if f.endswith('.avi') or f.endswith('.mp4')])
        # Get the file in each folder that ends with .csv
        csv_names = ([f for f in os.listdir(os.path.join(path, dirs)) if f.endswith('.csv')])

        if len(avi_names) == 1 and len(csv_names) == 1:
            new_csv_name = os.path.splitext(avi_names[0])[0] + '.csv'
            new_csv_names.extend(new_csv_name)
            os.rename(os.path.join(path, dirs, csv_names[0]), os.path.join(path, dirs, new_csv_name))
            folder_dict[os.path.splitext(avi_names[0])[0]] = os.path.join(path, dirs)
    return folder_dict


def update_main_cfg(main_path, new_dict):
    with open(main_path, 'r') as f:
        config = yaml.safe_load(f)

    config['experiment_data_paths'].update(new_dict)

    with open(main_path, 'w') as f:
        yaml.safe_dump(config, f)


def update_expt_info_df(df, yaml_path):
    """Updates the df containing experimental info with a supplemental configuration file that has the sex info
    for flies without it """
    with open(yaml_path, 'r') as f:
        info = yaml.safe_load(f)
    update_dict = info['sexlist']
    for key in list(update_dict.keys()):
        df['Sex'][df['ExptNames'] == key] = update_dict[key]
    return df


def get_likelihood(data_path_dict):
    """Loops through the data and loads likelihood scores for the desired body parts"""
    out_df_list = []
    for keys, values in data_path_dict.items():
        file_path = list(Path(values).glob(keys + '.csv'))[0]
        ind = 'likelihood'
        #save likelihood
        llh_path = os.path.join(values,keys + '_llh.pkl')
        if not os.path.exists(llh_path):
            # grab the header to create a new header for the df
            header = pd.read_csv(file_path, nrows=3)
            col_list = header.columns[header.loc[1] == ind]
            head_list = header[col_list].loc[0].tolist()
            df = pd.read_csv(file_path, skiprows=[0, 1, 2], header=None)
            df = df.iloc[:, 3::3]
            df.columns = head_list
            df.to_pickle(llh_path)
            df['ExptNames'] = keys
        else:
            df = pd.read_pickle(llh_path)
            df['ExptNames'] = keys
        out_df_list.append(df)

    out_df = pd.concat(out_df_list)
    return out_df


def convert_hour_to_HM(hour):
    return time.strftime("%H:%M", time.gmtime(float(hour) * 3600))


def change_points(labels):
    labels_diff = np.diff(labels)
    return (np.nonzero(labels_diff != 0)[0] + 1).astype(int)


def cont_intvls(labels):
    return np.append(
        np.insert(change_points(labels), 0, 0),
        labels.shape[0],
    ).astype(int)


def generate_bout_report(arr, label_to_name={}, filter_names=[]):
    bout_intervals = cont_intvls(arr)

    bout_start_list = []
    bout_finish_list = []
    name_list = []

    for i in range(bout_intervals.shape[0] - 1):
        start, finish = bout_intervals[i], bout_intervals[i + 1] - 1
        name = label_to_name[arr[bout_intervals[i]]]

        if not filter_names or name in filter_names:
            bout_start_list.append(start)
            bout_finish_list.append(finish)
            name_list.append(name)

    report_df = pd.DataFrame(
        {
            "Start": bout_start_list,
            "Finish": bout_finish_list,
            "Name": name_list,
        }
    )
    return report_df


def reverse_dict(x):
    if x is not None:
        reversed_dict = {j: i for i, j in x.items()}
    return reversed_dict


def inner_dict_sum(x):
    return {key: sum(val.values()) for key, val in x.items()}


def sliding_window(seq, winsize=3, stepsize=1):
    assert winsize > 1
    assert stepsize >= 1
    if winsize % 2 == 0:
        winsize = winsize + 1
    hw = (winsize - 1) // 2
    p1 = [seq[0: i + hw] for i in range(0, hw, stepsize)]
    p2 = [seq[i - hw: i + hw] for i in range(hw, len(seq), stepsize)]
    return p1 + p2


def most_common(lst):
    if not lst:
        return -1
    data = Counter(lst)
    return data.most_common(1)[0][0]


def flatten(items, seqtypes=(list, tuple)):
    for i, _ in enumerate(items):
        while i < len(items) and isinstance(items[i], seqtypes):
            items[i: i + 1] = items[i]
    return items


def multi_append(*args):
    for list_, item in args:
        list_.append(item)


def merge_two_dicts(x, y):
    z = x.copy()  # Start with x's keys and values.
    z.update(y)  # Modifies z with y's keys and values & returns None.
    return z


# Return the first nested list which contains the item.
# If there is no such list, return [].
def in_nested_list(my_list, item):
    is_in = []
    if item in my_list:
        is_in = my_list
    else:
        for sublist in my_list:
            if isinstance(sublist, list):
                is_in = in_nested_list(sublist, item)
            if len(is_in) > 0:
                break
    return is_in


def is_all_equal(iterator):
    iterator = iter(iterator)
    try:
        first = next(iterator)
    except StopIteration:
        return True
    return all(first == x for x in iterator)


def make_funcname_dict(*args):
    dict_ = {}
    for item in args:
        dict_.update({item.__name__: item})
    return dict_


def timeit(f):
    """
    Timing decorator for functions. Just add @timeit to start and function
    will be timed. It will write starting and ending times

    Parameters
    ----------
    f : function
        decorator takes the function as parameter
    Returns
    -------
    mixed
        return value of function itself
    Raises
    ------
    Error
        when any error thrown by called function does not catch it
    """

    def wrapper(*args, **kwargs):
        logger = logging.getLogger("main")
        logging.info("...")
        logging.basicConfig()
        logger.info(f"Started: {f.__qualname__}.")
        t = time.time()
        res = f(*args, **kwargs)
        logger.info(f"Finished: {f.__qualname__} elapsed: {time.time() - t:.2f}s.")
        logging.info("...")
        return res

    return wrapper



import warnings
from pathlib import Path

import numpy as np
from sklearn.preprocessing import robust_scale
from tqdm import tqdm

import basty.utils.io as io
import basty.utils.misc as misc
from basty.experiment_processing.experiment_info import ExptRecord
from basty.experiment_processing.experiment_outlining import ActiveBouts, DormantEpochs
from basty.project.helper import LoadingHelper, Logger, ParameterHandler, SavingHelper
from basty.utils.annotations import HumanAnnotations as HumAnn

warnings.filterwarnings("ignore")


class Project(ParameterHandler, LoadingHelper, SavingHelper):
    def __init__(self, main_cfg_path, **kwargs):
        self.logger = Logger(filepath=None)
        ParameterHandler.__init__(self)
        LoadingHelper.__init__(self, self.logger)
        SavingHelper.__init__(self, self.logger)

        self.main_cfg = io.read_config(main_cfg_path)

        def try_to_get_value(cfg, key, name):
            try:
                value = cfg[key]
            except KeyError as err:
                self.logger.missing_cfg_key_error(name)
                self.logger.direct_error(err)
                raise
            return value

        def try_to_get_path(cfg, key, name):
            value = Path(try_to_get_value(cfg, key, name))
            return value

        def try_to_read_cfg(path, name):
            try:
                cfg = io.read_config(path)
            except FileNotFoundError as err:
                self.logger.file_missing(name, path, depth=2)
                self.logger.direct_error(err)
                raise
            return cfg

        # Read main project path and create a directories at that path.
        self.project_path = try_to_get_path(
            self.main_cfg, "project_path", "project path"
        )
        io.safe_create_dir(self.project_path)

        # Read each configuration file paths given in the main configuration.
        self.cfg_path_dict = try_to_get_value(
            self.main_cfg, "configuration_paths", "configuration paths"
        )
        self.pose_cfg_path = try_to_get_path(
            self.cfg_path_dict, "pose_cfg", "pose configuration path"
        )
        self.feature_cfg_path = try_to_get_path(
            self.cfg_path_dict, "feature_cfg", "feature configuration path"
        )
        self.temporal_cfg_path = try_to_get_path(
            self.cfg_path_dict, "temporal_cfg", "temporal configuration path"
        )

        # This mode is to evaluate the pipeline and paramaters using annotated data.
        self.evaluation_mode = self.main_cfg.get("evaluation_mode", False)

        # Read and load each each configuration file.
        self.pose_cfg = try_to_read_cfg(self.pose_cfg_path, "pose configuration")
        self.feature_cfg = try_to_read_cfg(
            self.feature_cfg_path, "feature configuration"
        )
        self.temporal_cfg = try_to_read_cfg(
            self.temporal_cfg_path, "temporal configuration"
        )

        # Read FPS value from the temporal configuration.
        self.fps = try_to_get_value(self.temporal_cfg, "fps", "fps value")

        # Read experiment data paths from the main configuration.
        expt_data_paths = try_to_get_value(
            self.main_cfg, "experiment_data_paths", "experiment data paths"
        )
        self.data_path_dict = {key: Path(val) for key, val in expt_data_paths.items()}

        # Duplicate and update configuration files to use later.
        # Save them into the project directory at the given project path.
        self.main_cfg["configuration_paths"]["feature_cfg"] = str(
            self.project_path / "feature_cfg.yaml"
        )
        self.main_cfg["configuration_paths"]["pose_cfg"] = str(
            self.project_path / "pose_cfg.yaml"
        )
        self.main_cfg["configuration_paths"]["temporal_cfg"] = str(
            self.project_path / "temporal_cfg.yaml"
        )

        cfg_path_dict = self.main_cfg["configuration_paths"]
        io.dump_yaml(self.pose_cfg, cfg_path_dict["pose_cfg"])
        io.dump_yaml(self.feature_cfg, cfg_path_dict["feature_cfg"])
        io.dump_yaml(self.temporal_cfg, cfg_path_dict["temporal_cfg"])
        io.dump_yaml(self.main_cfg, self.project_path / "main_cfg.yaml")

        # Create sub-directories in the project for each experiment.
        self.expt_names = []
        self.expt_path_dict = {}

        for name, _ in self.data_path_dict.items():
            self.expt_names.append(name)

            expt_path = self.project_path / name
            self.expt_path_dict[name] = expt_path
            io.safe_create_dir(expt_path)
            io.safe_create_dir(expt_path / "embeddings")
            io.safe_create_dir(expt_path / "clusterings")
            io.safe_create_dir(expt_path / "correspondences")
            io.safe_create_dir(expt_path / "figures")

            if not (expt_path / "expt_record.z").exists():
                self.logger.direct_info(
                    f"Creating empty experiment record objects for {name}"
                )
                expt_record = ExptRecord(
                    name, self.data_path_dict[name], expt_path, fps=self.fps
                )
                self._save_joblib_object(expt_record, expt_path, "expt_record.z")
            # else:
            #     self.logger.file_already_exists(
            #         "expt_record.z", expt_path / "expt_record.z", depth=2
            #     )

        # Process, save and visualize given human annotations.
        self.init_annotation_kwargs(**kwargs)
        self.annotation_path_dict = self.main_cfg.get("annotation_paths", {})

        for idx, (expt_name, ann_path) in enumerate(self.annotation_path_dict.items()):
            expt_path = self.expt_path_dict[expt_name]

            if not (expt_path / "annotations.npy").exists():
                self.logger.direct_info(f"Processing human annotations of {expt_name}")

                annotator = HumAnn(
                    ann_path,
                    inactive_annotation=self.inactive_annotation,
                    noise_annotation=self.noise_annotation,
                    arouse_annotation=self.arouse_annotation,
                )

                if idx > 0:
                    assert annotator.behavior_to_label == expt_record.behavior_to_label
                    assert annotator.label_to_behavior == expt_record.label_to_behavior

                expt_record = self._load_joblib_object(expt_path, "expt_record.z")
                expt_record.has_annotation = True

                y_ann_list = annotator.get_annotations()
                y_ann = annotator.label_converter(
                    y_ann_list, priority_order=self.annotation_priority
                )
                self._save_numpy_array(y_ann, expt_path, "annotations.npy")

                expt_record.inactive_annotation = annotator.inactive_annotation
                expt_record.noise_annotation = annotator.noise_annotation
                expt_record.arouse_annotation = annotator.arouse_annotation
                expt_record.label_to_behavior = annotator.label_to_behavior
                expt_record.behavior_to_label = annotator.behavior_to_label
                expt_record.mask_noise = (
                    annotator.behavior_to_label[annotator.noise_annotation] == y_ann
                )
                mask_annotated = (
                    annotator.behavior_to_label[annotator.inactive_annotation] != y_ann
                )
                expt_record.mask_annotated = np.logical_and(
                    mask_annotated, np.logical_not(expt_record.mask_noise)
                )

                io.safe_create_dir(expt_path / "annotations")
                for behavior in expt_record.behavior_to_label.keys():
                    ann_report_df = misc.generate_bout_report(
                        y_ann,
                        label_to_name=expt_record.label_to_behavior,
                        filter_names=[behavior],
                    )
                    ann_report_df.to_csv(
                        expt_path / "annotations" / f"{expt_name}-{behavior}.csv"
                    )

                self._save_joblib_object(
                    expt_record,
                    expt_path,
                    "expt_record.z",
                    "with annotation information",
                )
            # else:
            #     self.logger.file_already_exists(
            #         "annotations.npy", expt_path / "annotations.npy", depth=2
            #     )


class ExptOutline:
    def get_training_data(
        self,
        X_expt_dict,
        training_expt_names,
        get_default_training_labels,
        mask_training_data,
    ):
        X_train_list = []
        y_train_list = []

        for ann_expt_name in training_expt_names:
            expt_path = self.expt_path_dict[ann_expt_name]
            expt_record = self._load_joblib_object(expt_path, "expt_record.z")
            y_ann = self._load_numpy_array(expt_path, "annotations.npy")

            X_train = X_expt_dict[ann_expt_name]
            y_train = np.zeros(y_ann.shape, dtype=int)

            if self.label_conversion_dict:
                for (
                    new_label,
                    behavior_list,
                ) in self.label_conversion_dict.items():
                    for behavior in behavior_list:
                        old_label = expt_record.behavior_to_label[behavior]
                        y_train[y_ann == old_label] = new_label
            else:
                y_train = get_default_training_labels(y_train, y_ann, expt_record)

            X_train, y_train = mask_training_data(X_train, y_train, expt_record)
            X_train_list.append(X_train)
            y_train_list.append(y_train)
        return X_train_list, y_train_list


class ExptDormantEpochs(Project, ExptOutline):
    def __init__(self, main_cfg_path, use_supervised_learning=False, **kwargs):
        Project.__init__(self, main_cfg_path, **kwargs)
        self.use_supervised_learning = use_supervised_learning
        self.init_dormant_epochs_kwargs(self.fps, use_supervised_learning, **kwargs)

    @misc.timeit
    def outline_dormant_epochs(self):
        X_expt_dict = dict()
        annotated_expt_names = list(self.annotation_path_dict.keys())
        all_expt_names = list(self.expt_path_dict.keys())

        if self.evaluation_mode:
            expt_names = annotated_expt_names
        else:
            expt_names = all_expt_names

        pbar = tqdm(expt_names)
        for expt_name in pbar:
            expt_path = self.expt_path_dict[expt_name]
            pbar.set_description(
                f"Computing feature values of {expt_name} for outlining dormant epochs"
            )

            expt_record = self._load_joblib_object(expt_path, "expt_record.z")
            delta_stft = self._load_pandas_dataframe(expt_path, "delta_stft.pkl")
            ftname_to_deltaft = self._load_yaml_dictionary(
                expt_path, "ftname_to_deltaft.yaml"
            )

            if not self.datums:
                datums = list(ftname_to_deltaft.values())
            else:
                datums = self.datums

            X = DormantEpochs.get_datums_values(
                delta_stft,
                misc.reverse_dict(ftname_to_deltaft),
                datums=datums,
                winsize=self.datums_winsize,
            ).reshape((-1, 1))

            if self.log_scale:
                X = np.log2(X + 1)
            if self.scale:
                X = robust_scale(X)

            X_expt_dict[expt_name] = X

        dormant_epochs = DormantEpochs()

        if self.use_supervised_learning and not self.evaluation_mode:
            self.logger.direct_info(
                "Training the classifier for dormant bouts with all."
            )
            training_expt_names = annotated_expt_names
            X_train_list, y_train_list = self.get_training_data(
                X_expt_dict,
                training_expt_names,
                DormantEpochs.get_default_training_labels,
                DormantEpochs.mask_training_data,
            )
            dormant_epochs.construct_dormant_epochs_classifier(
                X_train_list, y_train_list, **self.classifier_kwargs
            )

        pbar = tqdm(expt_names)
        for expt_name in pbar:
            expt_path = self.expt_path_dict[expt_name]
            pbar.set_description(f"Outlining dormant epochs of {expt_name}")

            X = X_expt_dict[expt_name]
            expt_record = self._load_joblib_object(expt_path, "expt_record.z")

            if self.use_supervised_learning:
                if self.evaluation_mode:
                    self.logger.direct_info(
                        "Training the classifier for dormant epochs bouts."
                    )
                    training_expt_names = annotated_expt_names
                    training_expt_names.remove(expt_name)
                    X_train_list, y_train_list = self.get_training_data(
                        X_expt_dict,
                        training_expt_names,
                        DormantEpochs.get_default_training_labels,
                        DormantEpochs.mask_training_data,
                    )
                    dormant_epochs.construct_dormant_epochs_classifier(
                        X_train_list, y_train_list, **self.classifier_kwargs
                    )
                mask_dormant = dormant_epochs.predict_dormant_epochs(
                    X,
                    min_dormant=self.min_dormant,
                    winsize=self.post_processing_winsize,
                    wintype=self.post_processing_wintype,
                )
            else:
                threshold = DormantEpochs.get_threshold(
                    X,
                    self.threshold_key,
                    self.num_gmm_comp,
                    self.threshold_idx,
                )
                mask_dormant, final_labels = DormantEpochs.compute_dormant_epochs(
                    X,
                    threshold,
                    self.min_dormant,
                    self.tol_duration,
                    self.tol_percent,
                    self.epoch_winsize,
                )

            expt_record.mask_dormant = mask_dormant
            assert expt_record.mask_dormant.any()

            self._save_joblib_object(
                expt_record,
                expt_path,
                "expt_record.z",
                "with a mask for dormant epochs",
            )


class ExptActiveBouts(Project, ExptOutline):
    def __init__(self, main_cfg_path, use_supervised_learning=False, **kwargs):
        Project.__init__(self, main_cfg_path, **kwargs)
        self.use_supervised_learning = use_supervised_learning
        self.init_active_bouts_kwargs(self.fps, use_supervised_learning, **kwargs)

    @misc.timeit
    def outline_active_bouts(self):
        X_expt_dict = dict()
        annotated_expt_names = list(self.annotation_path_dict.keys())
        all_expt_names = list(self.expt_path_dict.keys())

        if self.evaluation_mode:
            expt_names = annotated_expt_names
        else:
            expt_names = all_expt_names

        pbar = tqdm(expt_names)
        for expt_name in pbar:
            expt_path = self.expt_path_dict[expt_name]
            pbar.set_description(
                f"Computing feature values of {expt_name} for outlining active bouts"
            )

            expt_record = self._load_joblib_object(expt_path, "expt_record.z")
            wsft = self._load_numpy_array(expt_path, "wsft.npy")
            ftname_to_snapft = self._load_yaml_dictionary(
                expt_path, "ftname_to_snapft.yaml"
            )

            df_coefs = ActiveBouts.get_df_coefs_summary(
                wsft,
                ftname_to_snapft,
                log=self.log_scale,
                method=self.coefs_summary_method,
            )
            del wsft

            if not any(self.datums_list):
                self.datums_list = [
                    [ftname_to_snapft[datum]] for datum in df_coefs.columns
                ]  # [[]]
            else:
                self.datums_list = self.datums_list

            values = []
            for datums in self.datums_list:
                values.append(
                    ActiveBouts.get_datums_values(
                        df_coefs,
                        misc.reverse_dict(ftname_to_snapft),
                        datums=datums,
                        winsize=self.datums_winsize,
                    )
                )
            del df_coefs

            X = np.stack(values, axis=1)

            if self.scale:
                X = robust_scale(X)

            X_expt_dict[expt_name] = X
            del values

        active_bouts = ActiveBouts()

        if self.use_supervised_learning and not self.evaluation_mode:
            self.logger.direct_info(
                "Training the classifier for active bouts with all."
            )
            training_expt_names = annotated_expt_names
            X_train_list, y_train_list = self.get_training_data(
                X_expt_dict,
                training_expt_names,
                ActiveBouts.get_default_training_labels,
                ActiveBouts.mask_training_data,
            )
            active_bouts.construct_active_bouts_classifier(
                X_train_list, y_train_list, **self.classifier_kwargs
            )

        pbar = tqdm(expt_names)
        for expt_name in pbar:
            expt_path = self.expt_path_dict[expt_name]
            pbar.set_description(f"Outlining active bouts of {expt_name}")

            X = X_expt_dict[expt_name]
            expt_record = self._load_joblib_object(expt_path, "expt_record.z")

            if self.use_supervised_learning:
                if self.evaluation_mode:
                    self.logger.direct_info("Training the classifier for active bouts.")
                    training_expt_names = annotated_expt_names[:]
                    training_expt_names.remove(expt_name)
                    X_train_list, y_train_list = self.get_training_data(
                        X_expt_dict,
                        training_expt_names,
                        ActiveBouts.get_default_training_labels,
                        ActiveBouts.mask_training_data,
                    )
                    active_bouts.construct_active_bouts_classifier(
                        X_train_list, y_train_list, **self.classifier_kwargs
                    )
                mask_active = active_bouts.predict_active_bouts(
                    X,
                    winsize=self.post_processing_winsize,
                    wintype=self.post_processing_wintype,
                )
            else:
                thresholds = []
                for idx, datums in enumerate(self.datums_list):
                    thresholds.append(
                        ActiveBouts.get_threshold(
                            X[expt_record.mask_dormant, idx, np.newaxis],
                            self.threshold_key,
                            self.num_gmm_comp,
                            self.threshold_idx,
                        )
                    )
                mask_active = ActiveBouts.compute_active_bouts(
                    X,
                    thresholds,
                    winsize=self.post_processing_winsize,
                    wintype=self.post_processing_wintype,
                )

            expt_record.mask_active = mask_active
            if expt_record.has_annotation:
                expt_record.mask_active = np.logical_and(
                    expt_record.mask_active, np.logical_not(expt_record.mask_noise)
                )
            assert expt_record.mask_active.any()
            assert np.logical_and(
                expt_record.mask_active, expt_record.mask_dormant
            ).any()

            self._save_joblib_object(
                expt_record,
                expt_path,
                "expt_record.z",
                "with a mask for active bouts",
            )



import numpy as np
import pandas as pd
import scipy.ndimage.filters as filters
from sklearn.ensemble import RandomForestClassifier
from sklearn.mixture import GaussianMixture

import basty.utils.misc as misc
from basty.utils.postprocessing import PostProcessing

# from sklearn.neighbors import KNeighborsClassifier


class OutlineMixin(PostProcessing):
    @staticmethod
    def get_datums_values(df_values, stft_to_ftname, datums=[], winsize=3):
        assert isinstance(df_values, pd.DataFrame) and df_values.ndim == 2

        X = np.zeros(df_values.shape[0])
        for d in datums:
            if stft_to_ftname[d] in df_values.columns:
                X = X + df_values[stft_to_ftname[d]].to_numpy()
            else:
                raise ValueError(
                    f"Given datum name {stft_to_ftname[d]} is not defined."
                )

        X[np.isnan(X)] = np.inf

        if winsize > 1:
            X = filters.median_filter(X, winsize)

        return X


class OutlineClassifier(OutlineMixin):
    def __init__(self, **kwargs):
        # self.clf = KNeighborsClassifier(**kwargs)
        self.clf = RandomForestClassifier(**kwargs)
        self.is_fitted = False
        self.n_samples_per_class = {None: 0}

    def fit(self, X_train, y_train):
        self.clf.fit(X_train, y_train)
        self.n_samples = y_train.shape[0]
        self.n_samples_per_class = {
            label: count
            for label, count in zip(*np.unique(y_train, return_counts=True))
        }
        self.is_fitted = True

    def predict(self, X):
        # y_prob = self.clf.predict_proba(X)
        # for class_label, n_samples in self.n_samples_per_class.items():
        #     y_prob[:, class_label] /= (np.log2(n_samples+1)+1)
        # y = np.argmax(y_prob, axis=1)
        y = self.clf.predict(X)
        return y

    def construct_outline_classifier(self, X_train_list, y_train_list, **kwargs):
        assert isinstance(X_train_list, list) and isinstance(y_train_list, list)
        assert len(X_train_list) == len(y_train_list)
        assert all(
            isinstance(X_expt, np.ndarray) and X_expt.ndim == 2
            for X_expt in X_train_list
        )
        assert all(
            all(np.issubdtype(type(lbl), np.int) for lbl in y_train)
            for y_train in y_train_list
        )

        X_train = np.concatenate(X_train_list, axis=0)
        y_train = np.concatenate(y_train_list, axis=0)

        self.__init__(**kwargs)
        self.fit(X_train, y_train)
        return self.clf


class OutlineThreshold(OutlineMixin):
    @staticmethod
    def threshold_detection(X, **kwargs):
        assert isinstance(X, np.ndarray) and X.ndim == 2
        assert X.shape[1] == 1
        gmm = GaussianMixture(**kwargs)
        clusters = gmm.fit_predict(X)

        xc = np.stack((X[:, 0], clusters))
        xc_s = xc[:, xc[0, :].argsort()]
        X_s, clusters_s = xc_s[0, :], xc_s[1, :]

        cluster_boundaries = X_s[misc.change_points(clusters_s) + 1]

        # Choose appropriate one based on the distribution.
        return cluster_boundaries, sorted(list(map(lambda x: x[0], gmm.means_)))

    @classmethod
    def get_threshold(cls, X, threshold_key, num_gmm_comp, threshold_idx, **kwargs):
        assert threshold_idx < num_gmm_comp
        kwargs["n_components"] = num_gmm_comp
        diff_points, cluster_means = cls.threshold_detection(X, **kwargs)
        if threshold_key == "local_min":
            threshold = diff_points[threshold_idx]
        elif threshold_key == "local_max":
            threshold = cluster_means[threshold_idx]
        else:
            raise ValueError(
                f"Given threshold key {threshold_key} is not defined. "
                "Use (local_min) local minimum or (loca_max) local max."
            )
        return threshold


class SummaryCoefsCWT:
    @staticmethod
    def coef_logsum_scales(cwt_coefs):
        return np.log2(np.sum(cwt_coefs[:, :, :], axis=1) + 1)

    @staticmethod
    def coef_logmax_scales(cwt_coefs):
        return np.log2(np.amax(cwt_coefs[:, :, :], axis=1) + 1)

    @staticmethod
    def coef_sum_scales(cwt_coefs):
        return np.sum(cwt_coefs[:, :, :], axis=1)

    @staticmethod
    def coef_max_scales(cwt_coefs):
        return np.max(cwt_coefs[:, :, :], axis=1)

    @classmethod
    def get_df_coefs_summary(cls, cwt_coefs, colum_names, log=True, method="sum"):
        assert len(colum_names) == cwt_coefs.shape[-1]
        known_methods = ["sum", "max"]

        if log and method == "sum":
            df = pd.DataFrame(
                cls.coef_logsum_scales(cwt_coefs),
                columns=colum_names.keys(),
            )
        elif log and method == "max":
            df = pd.DataFrame(
                cls.coef_logmax_scales(cwt_coefs),
                columns=colum_names.keys(),
            )
        elif not log and method == "max":
            df = pd.DataFrame(
                cls.coef_max_scales(cwt_coefs),
                columns=colum_names.keys(),
            )
        elif not log and method == "sum":
            df = pd.DataFrame(
                cls.coef_sum_scales(cwt_coefs),
                columns=colum_names.keys(),
            )
        else:
            raise ValueError(f"Unkown 'method', must be one of {str(known_methods)}.")
        return df


class ActiveBouts(OutlineThreshold, OutlineClassifier, SummaryCoefsCWT):
    @classmethod
    def mask_training_data(cls, X_train, y_train, expt_record):
        X_train = X_train[expt_record.mask_dormant]
        y_train = y_train[expt_record.mask_dormant]
        return X_train, y_train

    @classmethod
    def get_default_training_labels(cls, y_train, y_ann, expt_record):
        inactive_label, noise_label = (
            expt_record.behavior_to_label[expt_record.inactive_annotation],
            expt_record.behavior_to_label[expt_record.noise_annotation],
        )
        y_train[y_ann != inactive_label] += 1
        y_train[y_ann == noise_label] += 1
        return y_train

    @classmethod
    def compute_active_bouts(cls, X, thresholds, winsize=30, wintype="boxcar"):
        assert isinstance(X, np.ndarray) and X.ndim == 2
        assert len(thresholds) == X.shape[1]

        mask_init = np.full(X.shape[0], False, dtype=np.bool_)

        for i in range(X.shape[1]):
            values_per_datums = X[:, i]
            mask_init = np.logical_or(mask_init, values_per_datums > thresholds[i])

        initial_labels = mask_init.astype(int)
        intermediate_labels = cls.compute_window_majority(
            initial_labels, winsize=winsize, wintype=wintype
        )
        final_labels = cls.postprocess_wrt_durations(
            intermediate_labels, [1], winsize // 2
        )
        mask_active = final_labels == 1
        return mask_active

    def construct_active_bouts_classifier(self, X_train_list, y_train_list, **kwargs):
        self.construct_outline_classifier(X_train_list, y_train_list, **kwargs)

    def predict_active_bouts(self, X, winsize=30, wintype="boxcar"):
        assert isinstance(X, np.ndarray) and X.ndim == 2

        initial_labels = np.array(self.predict(X))
        intermediate_labels = self.compute_window_majority(
            initial_labels, winsize=winsize, wintype=wintype
        )
        final_labels = self.__class__.postprocess_wrt_durations(
            intermediate_labels, [1], winsize // 2
        )
        mask_active = final_labels == 1
        return mask_active


class DormantEpochs(OutlineThreshold, OutlineClassifier):
    label_to_name = {0: "Dormant", 1: "Arouse", -1: "Betwixt"}

    @classmethod
    def mask_training_data(cls, X_train, y_train, expt_record):
        return X_train, y_train

    @classmethod
    def get_default_training_labels(cls, y_train, y_ann, expt_record):
        arouse_label, noise_label = (
            expt_record.behavior_to_label[expt_record.arouse_annotation],
            expt_record.behavior_to_label[expt_record.noise_annotation],
        )
        y_train[np.logical_or(y_ann == arouse_label, y_ann == noise_label)] += 1
        y_train[y_ann == noise_label] += 1
        return y_train

    @classmethod
    def compute_dormant_epochs(
        cls,
        X,
        threshold,
        min_dormant,
        tol_duration,
        tol_percent=0.4,
        winsize=30,
    ):
        assert isinstance(X, np.ndarray) and X.ndim == 2
        assert X.shape[1] == 1

        initial_labels = np.where(X[:, 0] < threshold, 0, 1).astype(int)
        cont_intvls_dict = cls.get_cont_intvsl_dict(initial_labels)
        label_win = misc.sliding_window(initial_labels, winsize=winsize, stepsize=1)
        intermediate_labels = np.empty_like(initial_labels)

        for i, (lbl, item) in enumerate(cont_intvls_dict.items()):
            if lbl == 0:
                for intvl_start, intvl_end in item:
                    intermediate_labels[intvl_start:intvl_end] = 0  # dormant
            else:
                for intvl_start, intvl_end in item:
                    dur = intvl_end - intvl_start
                    short_arouse = dur < tol_duration
                    epoch_mid = intvl_start + dur // 2
                    mostly_dormant = (
                        np.sum(label_win[epoch_mid]) / winsize
                    ) < tol_percent
                    if short_arouse and mostly_dormant:
                        intermediate_labels[intvl_start:intvl_end] = -1  # betwixt
                    else:
                        intermediate_labels[intvl_start:intvl_end] = 1  # arouse
        intermediate_labels[intermediate_labels == -1] = 0
        final_labels = cls.postprocess_wrt_durations(
            intermediate_labels, [0], min_dormant
        )
        mask_dormant = final_labels == 0
        return mask_dormant, final_labels

    def construct_dormant_epochs_classifier(self, X_train_list, y_train_list, **kwargs):
        self.construct_outline_classifier(X_train_list, y_train_list, **kwargs)

    def predict_dormant_epochs(self, X, min_dormant=900, winsize=90, wintype="boxcar"):
        assert isinstance(X, np.ndarray) and X.ndim == 2

        initial_labels = np.array(self.predict(X))
        intermediate_labels = self.compute_window_majority(
            initial_labels, winsize=winsize, wintype=wintype
        )
        final_labels = self.__class__.postprocess_wrt_durations(
            intermediate_labels, [0], min_dormant
        )
        mask_dormant = final_labels == 0
        return mask_dormant


class OutlineClustering(OutlineMixin):
    @classmethod
    def get_component(cls, X, num_gmm_comp, component_idx, **kwargs):
        assert isinstance(X, np.ndarray) and X.ndim == 2
        assert X.shape[1] == 1
        gmm = GaussianMixture(n_components=num_gmm_comp, **kwargs)
        components = gmm.fit_predict(X)

        assert component_idx >= 0 and component_idx < num_gmm_comp
        components_sorted_by_mean = sorted(
            [c for c in range(num_gmm_comp)], key=lambda c: gmm.means_[c][0]
        )
        mask_component = components == components_sorted_by_mean[component_idx]

        return mask_component



from collections import defaultdict

import numpy as np
import pandas as pd

import basty.utils.misc as misc


class AnnotationInfo:
    def __init__(self):
        self._inactive_annotation = None
        self._noise_annotation = None
        self._arouse_annotation = None
        self._label_to_behavior = None
        self._behavior_to_label = None
        self._mask_annotated = None
        # no annotation by default
        self.has_annotation = False
        self.use_annotations_to_mask = {}

    @property
    def noise_annotation(self):
        if self._noise_annotation is None:
            raise ValueError("'noise_annotation' is not set.")
        return self._noise_annotation

    @noise_annotation.setter
    def noise_annotation(self, value):
        if not isinstance(value, str):
            raise ValueError("'noise_annotation' must have type 'str'.")
        self._noise_annotation = value

    @property
    def inactive_annotation(self):
        if self._inactive_annotation is None:
            raise ValueError("'inactive_annotation' is not set.")
        return self._inactive_annotation

    @inactive_annotation.setter
    def inactive_annotation(self, value):
        if not isinstance(value, str):
            raise ValueError("'inactive_annotation' must have type 'str'.")
        self._inactive_annotation = value

    @property
    def arouse_annotation(self):
        if self._arouse_annotation is None:
            raise ValueError("'arouse_annotation' is not set.")
        return self._arouse_annotation

    @arouse_annotation.setter
    def arouse_annotation(self, value):
        if not isinstance(value, str):
            raise ValueError("'arouse_annotation' must have type 'str'.")
        self._arouse_annotation = value

    @property
    def mask_annotated(self):
        if self._mask_annotated is None:
            raise ValueError("'mask_annotated' is not set.")
        return self._mask_annotated

    @mask_annotated.setter
    def mask_annotated(self, value):
        if not isinstance(value, np.ndarray) and value.ndim == 1:
            raise ValueError("'mask_annotated' must have type '1D numpy.ndarray'.")
        self._mask_annotated = value

    @property
    def label_to_behavior(self):
        if self._label_to_behavior is None:
            raise ValueError("'label_to_behavior' is not set.")
        return self._label_to_behavior

    @property
    def behavior_to_label(self):
        if self._behavior_to_label is None:
            raise ValueError("'behavior_to_label' is not set.")
        return self._behavior_to_label

    @label_to_behavior.setter
    def label_to_behavior(self, value):
        if not isinstance(value, dict):
            raise ValueError("'label_to_behavior' must have type 'dict'.")
        self._label_to_behavior = misc.sort_dict(value)
        self._behavior_to_label = misc.reverse_dict(self._label_to_behavior)

    @behavior_to_label.setter
    def behavior_to_label(self, value):
        if not isinstance(value, dict):
            raise ValueError("'behavior_to_label' must have type 'dict'.")
        self._behavior_to_label = misc.sort_dict(value)
        self._label_to_behavior = misc.reverse_dict(self._behavior_to_label)


class ExptInfo:
    def __init__(self, fps):
        self.fps = fps
        self.expt_frame_count = None
        self._part_to_frame_count = None
        self._mask_dormant = None
        self._mask_active = None
        self._mask_noise = None
        self._ftname_to_snapft = None
        self._ftname_to_deltaft = None
        self._snapft_to_ftname = None
        self._deltaft_to_ftname = None

    @property
    def part_to_frame_count(self):
        if self._part_to_frame_count is None:
            raise ValueError("'part_to_frame_count' is not set.")
        return self._part_to_frame_count

    @part_to_frame_count.setter
    def part_to_frame_count(self, value):
        if not isinstance(value, dict):
            raise ValueError("'part_to_frame_count' must have type 'dict'.")
        self.expt_frame_count = sum(value.values())
        self._part_to_frame_count = value

    @property
    def mask_noise(self):
        if self._mask_noise is None:
            raise ValueError("'mask_noise' is not set.")
        return self._mask_noise

    @mask_noise.setter
    def mask_noise(self, value):
        if not isinstance(value, np.ndarray) and value.ndim == 1:
            raise ValueError("'mask_noise' must have type '1D numpy.ndarray'.")
        self._mask_noise = value

    @property
    def mask_dormant(self):
        if self._mask_dormant is None:
            raise ValueError("'mask_dormant' is not set.")
        return self._mask_dormant

    @mask_dormant.setter
    def mask_dormant(self, value):
        if not isinstance(value, np.ndarray) and value.ndim == 1:
            raise ValueError("'mask_dormant' must have type '1D numpy.ndarray'.")
        self._mask_dormant = value

    @property
    def mask_active(self):
        if self._mask_active is None:
            raise ValueError("'mask_active' is not set.")
        if self._mask_noise is not None:
            mask_active = np.logical_and(
                self._mask_active, np.logical_not(self._mask_noise)
            )
        else:
            mask_active = self._mask_active
        return mask_active

    @mask_active.setter
    def mask_active(self, value):
        if not isinstance(value, np.ndarray) and value.ndim == 1:
            raise ValueError("'mask_active' must have type '1D numpy.ndarray'.")
        self._mask_active = value

    @property
    def ftname_to_snapft(self):
        if self._ftname_to_snapft is None:
            raise ValueError("'ftname_to_snapft' is not set.")
        return self._ftname_to_snapft

    @property
    def snapft_to_ftname(self):
        if self._snapft_to_ftname is None:
            raise ValueError("'snapft_to_ftname' is not set.")
        return self._snapft_to_ftname

    @property
    def ftname_to_deltaft(self):
        if self._ftname_to_deltaft is None:
            raise ValueError("'ftname_to_deltaft' is not set.")
        return self._ftname_to_deltaft

    @property
    def deltaft_to_ftname(self):
        if self._deltaft_to_ftname is None:
            raise ValueError("'deltaft_to_ftname' is not set.")
        return self._deltaft_to_ftname

    @snapft_to_ftname.setter
    def snapft_to_ftname(self, value):
        if not isinstance(value, dict):
            raise ValueError("'snapft_to_ftname' must have type 'dict'.")
        self._snapft_to_ftname = misc.sort_dict(value)
        self._ftname_to_snapft = misc.reverse_dict(self._snapft_to_ftname)

    @ftname_to_snapft.setter
    def ftname_to_snapft(self, value):
        if not isinstance(value, dict):
            raise ValueError("'ftname_to_snapft' must have type 'dict'.")
        self._ftname_to_snapft = misc.sort_dict(value)
        self._snapft_to_ftname = misc.reverse_dict(self._ftname_to_snapft)

    @deltaft_to_ftname.setter
    def deltaft_to_ftname(self, value):
        if not isinstance(value, dict):
            raise ValueError("'deltaft_to_ftname' must have type 'dict'.")
        self._deltaft_to_ftname = misc.sort_dict(value)
        self._ftname_to_deltaft = misc.reverse_dict(self._deltaft_to_ftname)

    @ftname_to_deltaft.setter
    def ftname_to_deltaft(self, value):
        if not isinstance(value, dict):
            raise ValueError("'ftname_to_deltaft' must have type 'dict'.")
        self._ftname_to_deltaft = misc.sort_dict(value)
        self._deltaft_to_ftname = misc.reverse_dict(self._ftname_to_deltaft)

    def _get_sec(self, idx):
        sec_total = idx // self.fps
        second = sec_total % 60
        return sec_total, second

    def get_hour_stamp(self, idx):
        if idx is np.nan:
            stamp = np.nan
        else:
            sec_total, second = self._get_sec(idx)
            minute = (sec_total // 60) % 60
            hour = sec_total // 3600
            stamp = str(int(hour)) + ":" + str(int(minute)) + ":" + str(int(second))
        return stamp

    def get_minute_stamp(self, idx):
        if idx is np.nan:
            stamp = np.nan
        else:
            sec_total, second = self._get_sec(idx)
            minute = sec_total // 60
            stamp = str(int(minute)) + ":" + str(int(second))
        return stamp


class ExptRecord(ExptInfo, AnnotationInfo):
    def __init__(self, name, data_path, expt_path, fps=30):
        self.name = name
        self.data_path = data_path
        self.expt_path = expt_path
        ExptInfo.__init__(self, fps)
        AnnotationInfo.__init__(self)

    def generate_report(self, labels, is_behavior=False, use_time_stamps=False):
        assert isinstance(labels, np.ndarray)
        assert np.issubdtype(labels.dtype, np.integer)

        intvls = misc.cont_intvls(labels)

        if is_behavior:
            try:
                labels = [self.label_to_behavior[lbl] for lbl in labels]
            except KeyError:
                raise ValueError(
                    "Given labels are not defined for label to behavior mapping."
                )

        report_dict = defaultdict(list)
        for i in range(intvls.shape[0] - 1):
            dur = intvls[i + 1] - intvls[i]
            if use_time_stamps:
                report_dict["Duration"].append(self.get_minute_stamp(dur))
                report_dict["Beginning"].append(self.get_hour_stamp(intvls[i]))
                report_dict["End"].append(self.get_hour_stamp(intvls[i + 1]))
            else:
                report_dict["Duration"].append(dur)
                report_dict["Beginning"].append(intvls[i])
                report_dict["End"].append(intvls[i + 1])
            report_dict["Label"].append(labels[intvls[i]])
        df_report = pd.DataFrame.from_dict(report_dict)

        return df_report



import pathlib

import numpy as np
import pandas as pd

import basty.utils.misc as misc


class HumanAnnotations:
    def __init__(
        self,
        ann_path=None,
        inactive_annotation="Idle&Other",
        noise_annotation="Noise",
        arouse_annotation="Moving",
    ):
        # Label of the inactive behavior is always zero.
        self.inactive_annotation = inactive_annotation
        self.noise_annotation = noise_annotation
        self.arouse_annotation = arouse_annotation
        self.behavior_to_label = {}
        self.multibehavior_to_label = {}

        self.behavior_to_label[self.inactive_annotation] = 0
        self.multibehavior_to_label[self.inactive_annotation] = 0
        self.behavior_to_label[self.arouse_annotation] = 1
        self.multibehavior_to_label[self.arouse_annotation] = 1
        self.behavior_to_label[self.noise_annotation] = 2
        self.multibehavior_to_label[self.noise_annotation] = 2
        static_annotation_list = [
            self.inactive_annotation,
            self.arouse_annotation,
            self.noise_annotation,
        ]

        if ann_path is not None:
            self.df_ann = pd.read_csv(ann_path)
            behaviors_uniq = [
                behavior
                for behavior in self.df_ann["Behavior"].unique()
                if behavior not in static_annotation_list
            ]
            for i, bhv in enumerate(sorted(behaviors_uniq)):
                self.behavior_to_label[bhv] = i + 3
                self.multibehavior_to_label[bhv] = i + 3
            self.df_ann = self.df_ann.query("Beginning != End")
        else:
            self.df_ann = pd.DataFrame()
            behaviors_uniq = []

        self.label_to_behavior = misc.reverse_dict(self.behavior_to_label)
        self.label_to_multibehavior = misc.reverse_dict(self.multibehavior_to_label)

    def annotation_details(self, y_ann):
        assert isinstance(y_ann, np.ndarray)
        ann_uniqs, ann_counts = np.unique(y_ann, return_counts=True)
        detal_dict = {}
        for idx, lbl_ann in enumerate(ann_uniqs):
            try:
                detal_dict[self.label_to_behavior[lbl_ann]] = ann_counts[idx]
            except KeyError:
                try:
                    detal_dict[self.label_to_multibehavior[lbl_ann]] = ann_counts[idx]
                except KeyError:
                    raise KeyError("Given annotation label is unkown.")
        return detal_dict

    def compile_annotations(self, ann_dir, ann_out_path, num_of_frames):
        annotations = {}
        missing_annotations = {}
        for path in pathlib.Path(ann_dir).glob("*.csv"):
            try:
                bhv = "-".join(path.stem.split("-")[-1].split("_"))
                df_tmp = pd.read_csv(path, header=None)
                df_tmp = df_tmp.rename(columns={0: "Beginning", 1: "End"})
                df_tmp = df_tmp.assign(
                    Behavior=pd.Series([bhv for _ in range(df_tmp.shape[0])]).values
                )
                annotations[bhv] = df_tmp
            except pd.errors.EmptyDataError:
                df_tmp = pd.DataFrame.from_dict(
                    {"Beginning": [-1], "End": [-1], "Behavior": [bhv]}
                )
                missing_annotations[bhv] = df_tmp

        df_ann = (
            pd.concat(annotations.values())
            .sort_values("Beginning")
            .reset_index(drop=True)
        )
        ann_stop_dict = {"Beginning": [], "End": [], "Behavior": []}
        for i in range(df_ann.shape[0]):
            ann_stop_dict["Behavior"].append(self.inactive_annotation)
            ann_stop_dict["Beginning"].append(df_ann["End"].iloc[i])
            if i == df_ann.shape[0] - 1:
                ann_stop_dict["End"].append(num_of_frames)
            else:
                ann_stop_dict["End"].append(df_ann["Beginning"].iloc[i + 1])

        if df_ann["Beginning"].iloc[0] != 0:
            ann_stop_dict["Behavior"].append(self.inactive_annotation)
            ann_stop_dict["Beginning"].append(0)
            ann_stop_dict["End"].append(df_ann["Beginning"].iloc[0])

        if missing_annotations:
            df_missing_ann = pd.concat(missing_annotations.values()).reset_index(
                drop=True
            )
        else:
            df_missing_ann = pd.DataFrame()
        df_ann = (
            pd.concat((df_missing_ann, df_ann, pd.DataFrame.from_dict(ann_stop_dict)))
            .sort_values("Beginning")
            .reset_index(drop=True)
        )
        df_ann.to_csv(ann_out_path)

    def get_annotations(self):
        y_ann_list = [[] for _ in range(self.df_ann["End"].max())]
        for i in range(self.df_ann.shape[0]):
            ann_name = self.df_ann.iloc[i]["Behavior"]
            begining = self.df_ann.iloc[i]["Beginning"]
            end = self.df_ann.iloc[i]["End"]
            for i in range(begining, end):
                y_ann_list[i].append(self.behavior_to_label[ann_name])
        return y_ann_list

    def label_converter(self, y_ann_list, priority_order=[]):
        def get_label(y_ann_i):
            assert len(y_ann_i) > 0
            lbl_ann = y_ann_i[-1]
            for ann_name in priority_order:
                lbl = self.behavior_to_label[ann_name]
                if lbl in y_ann_i:
                    lbl_ann = lbl
                    break
            return lbl_ann

        y_ann = np.array(
            [get_label(y_ann_list[i]) for i in range(len(y_ann_list))], dtype=int
        )
        return y_ann

    def multilabel_converter(self, y_ann_list, sep="&"):
        def get_label(y_ann_i):
            assert len(y_ann_i) > 0
            if len(y_ann_i) > 1:
                name_ann = sep.join(
                    [
                        self.label_to_multibehavior[y_ann_i[j]]
                        for j in range(len(y_ann_i))
                        if y_ann_i[j]
                        != self.behavior_to_label[self.inactive_annotation]
                    ]
                )
                lbl_ann = self.multibehavior_to_label.get(name_ann, False)
                if lbl_ann:
                    pass
                else:
                    lbl_ann = max(self.multibehavior_to_label.values()) + 1
                    self.multibehavior_to_label[name_ann] = lbl_ann
                    self.label_to_multibehavior[lbl_ann] = name_ann
            else:
                lbl_ann = y_ann_i[0]
            return lbl_ann

        y_ann = np.array(
            [get_label(y_ann_list[i]) for i in range(len(y_ann_list))], dtype=int
        )
        return y_ann



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['basty',
 'basty.behavior_mapping',
 'basty.experiment_processing',
 'basty.feature_extraction',
 'basty.project',
 'basty.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyWavelets>=1.2.0,<2.0.0',
 'PyYAML==6.0',
 'filterpy==1.4.5',
 'hdbscan==0.8.28',
 'joblib==1.1.0',
 'numpy==1.21.5',
 'pandas==1.3.5',
 'scikit-learn==1.1.1',
 'scipy==1.8.0',
 'tqdm==4.63.0',
 'umap-learn==0.5.2']

setup_kwargs = {
    'name': 'basty',
    'version': '0.1.0',
    'description': '',
    'long_description': 'None',
    'author': 'Ali Osman Berk Şapcı',
    'author_email': 'aliosmanberk@sabanciuniv.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '==3.9.0',
}


setup(**setup_kwargs)



import logging

import numpy as np
import pandas as pd
import scipy.signal as signal
from filterpy.kalman import KalmanFilter
from scipy.interpolate import splev, splrep
from scipy.ndimage import median_filter as medfilt
from scipy.stats import zscore

# This functions are only for coordinate values.
# Methods should be applied to pose dataframe directly.
# Filters are applied for each cartesian compenent of each body-part.


def rts_smoother_one_dimensional(df_pose, kalman_filter_kwargs, dim_x=2, dim_z=1):
    fk = KalmanFilter(dim_x=dim_x, dim_z=dim_z)
    # (dim_x, dim_x) process uncertainty, default eye(dim_x)
    fk.Q = kalman_filter_kwargs.get("process_uncertainty", np.eye(dim_x))
    # (dim_x, dim_x) state transistion matrix
    fk.F = kalman_filter_kwargs.get("state_transmission_matrix", np.eye(dim_x))
    # (dim_z, dim_x) measurement function
    fk.H = kalman_filter_kwargs.get("measurement_function", np.ones((dim_x, dim_z)))
    # (dim_x, dim_x) covariance matrix, default eye(dim_x)
    fk.P = kalman_filter_kwargs.get("covariance_matrix", np.eye(dim_x))
    # (dim_z, dim_z) measurement noise, default eye(dim_x)
    fk.R = kalman_filter_kwargs.get("noise", np.eye(dim_z))

    batch_size = kalman_filter_kwargs.get("batch_size", -1)
    if batch_size < 2:
        batch_size = df_pose.shape[0]
    t = np.arange(0, df_pose.shape[0])

    for col in df_pose.columns:
        pos = df_pose[col].to_numpy()
        f = splrep(t, pos, k=5, s=3)
        vel = splev(t, f, der=1) / 10
        for i in range(0, pos.shape[0], batch_size):
            # (dim_x, 1) initial location and velocity
            fk.x = np.array([[pos[i]], [vel[i]]])
            mu, cov, _, _ = fk.batch_filter(pos[i : i + batch_size])
            M, _, _, _ = fk.rts_smoother(mu, cov)
            pos[i : i + batch_size] = np.squeeze(M[:, 0])
        df_pose[col] = pos
    return df_pose


def rts_smoother_two_dimensional(df_pose):
    raise NotImplementedError


def boxcar_center_filter(df_pose, winsize):
    for name in df_pose.columns:
        df_pose[name] = (
            df_pose[name].rolling(window=winsize, min_periods=1, center=True).mean()
        )
    return df_pose


def median_filter(df_pose, winsize):
    for name in df_pose.columns:
        df_pose[name] = medfilt(df_pose[name].to_numpy(), size=winsize)
    return df_pose


def decimate(df_pose, q, ftype="iir"):
    decimated_dict = {}
    for name in df_pose.columns:
        decimated_dict[name] = signal.decimate(df_pose[name].to_numpy(), q, ftype=ftype)
    return pd.DataFrame.from_dict(decimated_dict)


def remove_decrasing_llh_frames(df_pose, df_llh, winsize=45, lower_z=-3):
    assert lower_z < 0
    pfilt = {}
    if winsize % 2 == 0:
        winsize += 1
    for col in df_llh.columns:
        llh = df_llh[col].astype(float).to_numpy()
        grad_llh = signal.savgol_filter(
            llh, window_length=winsize, polyorder=2, deriv=1
        )
        z = zscore(grad_llh, axis=0, ddof=0, nan_policy="omit")

        decreasing_idx = np.nonzero(z < lower_z)[0]
        pfilt[col] = round(decreasing_idx.shape[0] / llh.shape[0], 5)

        df_pose[col + "_x"].iloc[decreasing_idx] = np.nan
        df_pose[col + "_y"].iloc[decreasing_idx] = np.nan

    return (df_pose, pfilt)


def remove_low_llh_frames(df_pose, df_llh, threshold=0.15):
    pfilt = {}
    for col in df_llh.columns:
        llh = df_llh[col].astype(float)

        low_idx = np.nonzero(llh.to_numpy() < threshold)[0]
        pfilt[col] = round(low_idx.shape[0] / llh.shape[0], 5)

        df_pose[col + "_x"].iloc[low_idx] = np.nan
        df_pose[col + "_y"].iloc[low_idx] = np.nan

    return (df_pose, pfilt)


# Mark jumps as NaN based on the gradient quantiles.
def remove_jumps(df_pose, q=0.999):
    grad = np.gradient(df_pose.to_numpy(), axis=0, edge_order=2)
    quan_grad = np.quantile(grad, q, axis=0)
    columns = df_pose.columns
    num_interpolated = 0

    for i in range(quan_grad.shape[0]):
        idx_nan = np.nonzero(grad[:, i] > quan_grad[i])[0]
        num_interpolated += len(idx_nan)
        df_pose[columns[i]].iloc[idx_nan] = np.nan

    avgnum_interpolated = round(num_interpolated / quan_grad.shape[0], 2)
    logger = logging.getLogger("main")
    logger.info(f"avg. number of frames removed as jump: {avgnum_interpolated}.")
    return df_pose, avgnum_interpolated


# This is for fun, terribly slow and unnecessary.
def iteratively_impute_jumps(df_pose, q=0.999):
    df_pose, _ = remove_jumps(df_pose, q)

    from sklearn.experimental import enable_iterative_imputer  # noqa: F401
    from sklearn.impute import IterativeImputer

    imputer = IterativeImputer(random_state=0)
    df_pose = imputer.fit_transform(df_pose.to_numpy())
    return df_pose


def remove_local_outliers(df_pose, winsize=15, threshold=9):
    for name in df_pose.columns:
        medfilt_pose = medfilt(df_pose[name].to_numpy(), size=winsize)
        df_pose[name].iloc[np.abs(df_pose[name] - medfilt_pose) > winsize] = np.nan
    return df_pose


def fill_interpolate(df_pose):
    df_pose = df_pose.interpolate(method="ffill", axis=0)
    df_pose = df_pose.interpolate(method="bfill", axis=0)
    df_pose = df_pose.interpolate(method="linear", axis=0)
    return df_pose


def forward_impute_jumps(df_pose):
    raise NotImplementedError



import logging
import math
from copy import deepcopy

import numpy as np
import pywt


class WaveletTransformation:
    def __init__(self, fps, wv_cfg={}):
        self.wv_cfg = deepcopy(wv_cfg)
        self.logger = logging.getLogger("main")

        assert fps > 0
        self.sampling_freq = fps

        self.wavelet = self.wv_cfg.get("wavelet", "morl")
        self.padding = self.wv_cfg.get("padding", "reflect")
        self.num_channels = self.wv_cfg.get("num_channels", 20)
        self.freq_low = self.wv_cfg.get("freq_low", 0.5)
        self.freq_up = self.wv_cfg.get("freq_up", self.sampling_freq / 2)
        self.freq_spacing = self.wv_cfg.get("freq_spacing", "dyadically")
        self.normalization_option = self.wv_cfg.get("normalization_option", 2)

        assert self.num_channels > 1
        assert self.freq_low > 0
        assert self.freq_spacing in ["dyadically", "linearly"]

        if self.freq_up > (self.sampling_freq / 2):
            self.logger.warning(
                f"Upper bound frequency {self.freq_up} is greater than Nyquist frequency."
            )
            self.logger.info(
                f"Upper frequency is set to {round(self.sampling_freq/2, 3)}."
            )
            self.freq_up = self.sampling_freq / 2

        self.w0 = 5
        self.dc = 0.8125
        # In terms of Hz.
        self.dt = 1 / self.sampling_freq
        self.scales, self.freq_range = self._get_scales()

    def _get_scales(self):
        if self.wavelet == "morl":
            if self.freq_spacing == "dyadically":
                exp_f = math.log2(self.freq_up / self.freq_low)
                freq_range = self.freq_up * np.exp2(
                    np.arange(0, self.num_channels) * exp_f / (-self.num_channels + 1)
                )
                scales = self.dc * 1 / (freq_range * self.dt)
                # This is also possible, precision is worse though.
                # scales = (self.w0 + math.sqrt(2+self.w0**2)) / (4*math.pi*freq_range*self.dt)
                # freq_range = pywt.scale2frequency(self.wavelet, scales) / self.dt
            if self.freq_spacing == "linearly":
                freq_range = np.linspace(self.freq_low, self.freq_up, self.num_channels)
                scales = self.dc * 1 / (freq_range * self.dt)
        else:
            raise NotImplementedError(
                "Given wavelet {self.wavelet} is not implemented yet."
            )
        return scales, freq_range

    def cwtransform(self, feature_array):
        time_shape = feature_array.shape[0]
        scale_shape = self.scales.shape[0]
        feature_shape = feature_array.shape[1]

        # Using entire time axis as pad-width might be slow.
        pad_width = time_shape
        cwt_coefs = np.ndarray((time_shape, scale_shape, feature_shape))

        for i in range(feature_shape):
            padded = pywt.pad(feature_array[:, i], pad_width, self.padding)
            cws = pywt.cwt(padded, self.scales, self.wavelet, self.dt)
            ws = cws[0][:, time_shape : time_shape + pad_width]
            cwt_coefs[:, :, i] = np.swapaxes(ws, 0, 1)

        return cwt_coefs

    def normalize_channels(self, cwt_coefs):
        """

        Normalizes each time scale according to given option.

        normalization_option == 0
            pass
        normalization_option == 1
            Berman, Choi, Bialek, Shaevitz, (2014),
            Mapping the stereotyped behaviour of freely moving fruit flies,
            Journal of The Royal Society Interface.
        normalization_option == 2
            Liu, Y., San Liang, X., & Weisberg, R. H. (2007),
            Rectification of the Bias in the Wavelet Power Spectrum,
            Journal of Atmospheric and Oceanic Technology.

        """
        opt = self.normalization_option
        cwt_coefs_n = deepcopy(cwt_coefs)
        if opt == 0:
            pass
        elif self.wavelet == "morl" and opt == 1:
            exp1 = (math.pi ** (-0.25)) / np.sqrt(2 * self.scales)
            exp2 = math.exp((1 / 4) * (self.w0 - math.sqrt(self.w0**2 + 2)) ** 2)
            C = exp1 * exp2
            cwt_coefs_n[:, :, :] = np.abs(
                np.divide(cwt_coefs_n[:, :, :], C[np.newaxis, :, np.newaxis])
            )
        elif opt == 2:
            for i in range(cwt_coefs_n.shape[2]):
                for j in range(cwt_coefs_n.shape[1]):
                    s = self.scales[j]
                    cwt_coefs_n[:, j, i] = ((cwt_coefs_n[:, j, i]) ** 2) / s
        else:
            raise ValueError(
                f"For {self.wavelet}, unkown option {opt} is given for normalization."
            )
        return cwt_coefs_n



import logging
import pickle
from pathlib import Path

import yaml

import basty.utils.misc as misc


def safe_create_dir(d: Path):
    if not d.exists():
        logger = logging.getLogger("main")
        logger.info(f"Directory is not found, creating: {str(d)}")
        d.mkdir(parents=True)


def ensure_file_dir(file_path: Path):
    safe_create_dir(file_path.parent)


def read_config(file_path):
    file_path = Path(file_path)
    file_extension = file_path.suffix
    if file_extension in [".yaml", ".yml"]:
        return read_yaml(file_path)
    else:
        raise ValueError("Unkown configuration file type!")


def read_pickle(file_path, **kwargs):
    with open(str(file_path), "rb") as file:
        return pickle.read(file)


def write_pickle(data, file_path, **kwargs):
    with open(str(file_path), "wb") as file:
        pickle.dump(data, file, protocol=pickle.HIGHEST_PROTOCOL)


def write_txt(data, file_path, **kwargs):
    with open(str(file_path), "w", encoding="utf-8") as file:
        file.write(data)


def read_txt(file_path, **kwargs):
    with open(str(file_path), "r", encoding="utf-8") as file:
        return file.read()


def read_yaml(path, **kwargs):
    with open(str(path), "r") as file:
        return yaml.load(file, Loader=yaml.SafeLoader)


def dump_yaml(data, file_path, **kwargs):
    with open(str(file_path), "w") as file:
        yaml.dump(data, file, default_flow_style=False, sort_keys=False)


def multi_read(destination_dir, names, function):
    read_func = read_reference_dict[function]
    data_all = []
    destination_dir = Path(destination_dir)
    for i in range(names):
        file_path = destination_dir / names[i]
        data_all.append(read_func(file_path))
    return data_all


def multi_write(destination_dir, names, data_all, function):
    if len(names) != len(data_all):
        raise ValueError("Lengt of given name list and data list are not same!")
    destination_dir = Path(destination_dir)
    write_func = write_reference_dict[function]
    for i in range(names):
        file_path = destination_dir / names[i]
        write_func(file_path, data_all[i])


read_reference_dict = misc.make_funcname_dict(
    read_pickle,
    read_txt,
    read_yaml,
)
write_reference_dict = misc.make_funcname_dict(
    write_pickle,
    write_txt,
    dump_yaml,
)



from collections import defaultdict

import numpy as np
import sklearn.metrics as metrics
from sklearn.neighbors import NearestNeighbors

import basty.utils.misc as misc


def label_coverage(y_ann, y_pred, normalize=False):
    assert y_ann.shape[0] == y_pred.shape[0]

    pred_coverage = {}
    for y_hat in np.unique(y_pred):
        uniq_y_ann, counts = np.unique(y_ann[y_pred == y_hat], return_counts=True)
        coverage = (
            counts / (y_pred[y_pred == y_hat].shape[0] + 1) if normalize else counts
        )
        pred_coverage[y_hat] = dict(zip(uniq_y_ann, coverage))

    ann_coverage = {}
    for y_hat in np.unique(y_ann):
        uniq_y_pred, counts = np.unique(y_pred[y_ann == y_hat], return_counts=True)
        coverage = (
            counts / (y_ann[y_ann == y_hat].shape[0] + 1) if normalize else counts
        )
        ann_coverage[y_hat] = dict(zip(uniq_y_pred, coverage))

    return misc.sort_dict(ann_coverage), misc.sort_dict(pred_coverage)


def total_occupation(y):
    uniq_y, counts = np.unique(y, return_counts=True)
    return dict(zip(uniq_y.astype(int), counts))


def absolute_freq(y):
    occupation = total_occupation(y)
    return {key: val / y.shape[0] for key, val in occupation.items()}


def evolution_rate(y, normalize=False):
    uniq_y = np.unique(y)
    if normalize:
        denom = total_occupation(y)
    else:
        denom = [1 for _ in uniq_y]
    rate = {
        int(y_hat): np.cumsum(np.where(y == y_hat, 1, 0)) / denom[y_hat]
        for y_hat in uniq_y
    }
    return rate


class BoutDetails:
    def __init__(self, fps, y):
        self.fps = fps
        self.y = y
        self.intervals = misc.cont_intvls(y)

    def get_bout_feature_details(self, values, offset=1):
        intervals = self.intervals
        y = self.y
        bout_details_dict = defaultdict(list)

        for i in range(1, intervals.shape[0]):
            start = max(intervals[i - 1] - offset * self.fps, 0)
            end = intervals[i] + offset * self.fps
            bout_details_dict[y[intervals[i - 1]]].append(
                ((start, end), values[start:end, :])
            )
        bout_details_dict = misc.sort_dict(bout_details_dict)

        return bout_details_dict

    def get_bout_temporal_details(self):
        intervals = self.intervals
        y = self.y
        bout_details_dict = defaultdict(list)

        for i in range(1, intervals.shape[0]):
            start, end = intervals[i - 1], intervals[i]
            details = {"duration": end - start, "start": start, "end": end}
            details["label"] = y[intervals[i - 1]]
            bout_details_dict[y[intervals[i - 1]]].append(details)
        bout_details_dict = misc.sort_dict(bout_details_dict)

        return bout_details_dict


class kNN:
    def __init__(self, k):
        self.k = k
        self.kNeigh = NearestNeighbors(n_neighbors=self.k)

    def fit(self, X_embedded):
        self.kNeigh.fit(X_embedded)
        self.neighbours = self.kNeigh.kneighbors(X_embedded)

    def knn_accuracy(self, y, ignore_consecutive=0):
        assert ignore_consecutive >= 0
        neig_idx = self.neighbours[1]

        neig_avgdist = [
            sum([(i - j) for j in neig_idx[i]]) / self.k for i in range(len(neig_idx))
        ]
        majority_pred = [
            misc.most_common(
                [y[j] for j in neig_idx[i] if abs(j - i) > ignore_consecutive]
            )
            for i in range(len(neig_idx))
        ]

        accuracy = metrics.accuracy_score(y, majority_pred)
        print(f"Majority {self.k}-NN prediction accuracy is {accuracy}")

        return accuracy, neig_avgdist


class MetricMixin:
    def __init__(self, y_ann, y_pred, ignore_ann=None, ignore_pred=None):
        assert isinstance(y_ann, np.ndarray)
        assert isinstance(y_pred, np.ndarray)

        self.ignore_ann = [] if ignore_ann is None else ignore_ann
        self.ignore_pred = [] if ignore_pred is None else ignore_pred

        y_pred, y_ann = self.__class__._delete_y(y_pred, y_ann, ignore_pred)
        self.y_ann, self.y_pred = self.__class__._delete_y(y_ann, y_pred, ignore_ann)

    @staticmethod
    def _delete_y(y0, y1, ignore0):
        for y_hat in ignore0:
            y1 = y1[y0 != y_hat]
            y0 = y0[y0 != y_hat]
        return y0, y1


class ClassificationMetrics(MetricMixin):
    def __init__(self, y_ann, y_pred, ignore_ann=None, ignore_pred=None):
        super().__init__(y_ann, y_pred, ignore_ann=ignore_ann, ignore_pred=ignore_pred)
        self.ann_coverage, self.pred_coverage = label_coverage(
            self.y_ann, self.y_pred, normalize=True
        )
        pred_majority_dict = {
            y_hat: max(counts, key=(lambda key: counts[key]))
            for y_hat, counts in self.pred_coverage.items()
        }
        self.y_majority = np.array([pred_majority_dict[y_i] for y_i in self.y_pred])

        from sklearn import preprocessing

        lb = preprocessing.LabelBinarizer()
        self.y_ann_binary = lb.fit_transform(self.y_ann)

        self.y_pred_coverage_scores = np.zeros(self.y_ann_binary.shape)
        for i in range(self.y_pred_coverage_scores.shape[0]):
            for y_hat, val in enumerate(self.pred_coverage[self.y_pred[i]].items()):
                self.y_pred_coverage_scores[i, y_hat] = val

    def f1_score(self, **kwargs):
        return metrics.f1_score(self.y_ann, self.y_majority, **kwargs)

    def accuracy_score(self, **kwargs):
        return metrics.accuracy_score(self.y_ann, self.y_majority, **kwargs)

    def balanced_accuracy_score(self, **kwargs):
        return metrics.balanced_accuracy_score(self.y_ann, self.y_majority, **kwargs)

    def brier_score_loss(self, **kwargs):
        return metrics.brier_score_loss(self.y_ann, self.y_majority, **kwargs)

    def jaccard_score(self, **kwargs):
        return metrics.jaccard_score(self.y_ann, self.y_majority, **kwargs)

    def recall_score(self, **kwargs):
        return metrics.recall_score(self.y_ann, self.y_majority, **kwargs)

    def precision_score(self, **kwargs):
        return metrics.precision_score(self.y_ann, self.y_majority, **kwargs)

    def coverage_error(self, y_scores=None, **kwargs):
        y_scores = self.y_pred_coverage_scores if y_scores is None else y_scores
        return metrics.coverage_error(self.y_ann_binary, y_scores, **kwargs)

    def average_precision_score(self, y_scores=None, **kwargs):
        y_scores = self.y_pred_coverage_scores if y_scores is None else y_scores
        return metrics.average_precision_score(self.y_ann_binary, y_scores, **kwargs)

    def label_ranking_avg_precision_score(self, y_scores=None, **kwargs):
        y_scores = self.y_pred_coverage_scores if y_scores is None else y_scores
        return metrics.label_ranking_average_precision_score(
            self.y_ann_binary, y_scores, **kwargs
        )

    def label_ranking_loss(self, y_scores=None, **kwargs):
        y_scores = self.y_pred_coverage_scores if y_scores is None else y_scores
        return metrics.label_ranking_loss(self.y_ann_binary, y_scores, **kwargs)


class ClusteringMetrics(MetricMixin):
    def __init__(self, y_ann, y_pred, ignore_ann=None, ignore_pred=None):
        super().__init__(y_ann, y_pred, ignore_ann=ignore_ann, ignore_pred=ignore_pred)

    def adjusted_rand_score(self):
        (tn, fp), (fn, tp) = metrics.cluster.pair_confusion_matrix(
            self.y_ann, self.y_pred
        )
        # Special cases, namely empty data or full agreement.
        if fn == 0 and fp == 0:
            return 1.0
        return (
            2.0 * (tp * tn - fn * fp) / ((tp + fn) * (fn + tn) + (tp + fp) * (fp + tn))
        )

    def adjusted_mutual_info_score(self, **kwargs):
        return metrics.cluster.adjusted_mutual_info_score(
            self.y_ann, self.y_pred, **kwargs
        )

    def normalized_mutual_info_score(self, **kwargs):
        return metrics.cluster.normalized_mutual_info_score(
            self.y_ann, self.y_pred, **kwargs
        )

    def v_measure_score(self, **kwargs):
        return metrics.cluster.v_measure_score(self.y_ann, self.y_pred, **kwargs)

    def completeness_score(self):
        return metrics.cluster.completeness_score(self.y_ann, self.y_pred)

    def homogeneity_score(self):
        return metrics.cluster.homogeneity_score(self.y_ann, self.y_pred)


def compute_hausdorff_distance(X, y):
    # '0' stands for 'inactive' or 'unannotated' frames.
    assert 0 in y
    uniq_y = np.unique(y)
    # How far two subsets of a metric space are from each other.
    hausdorff_dist = np.zeros((uniq_y.shape[0], uniq_y.shape[0]))
    # hausdorff_dist[np.diag_indices(hausdorff_dist.shape[0])] = 0

    from scipy.spatial.distance import directed_hausdorff

    for i in range(1, hausdorff_dist.shape[0]):
        for j in range(i + 1, hausdorff_dist.shape[1]):
            dir1 = directed_hausdorff(X[y == uniq_y[i]], X[y == uniq_y[j]])
            dir2 = directed_hausdorff(X[y == uniq_y[i]], X[y == uniq_y[j]])
            hausdorff_dist[i, j] = (dir1[0] + dir2[0]) / 2
    hausdorff_dist = hausdorff_dist + np.triu(hausdorff_dist).T
    return hausdorff_dist


def compute_distn_jensenshannon(X, y):
    # '0' stands for 'inactive' or 'unannotated' frames.
    assert 0 in y
    uniq_y = np.unique(y)
    # Jensen-Shannon distance (metric) between two probability arrays of same size.
    js_dist = np.zeros((uniq_y.shape[0], uniq_y.shape[0]))

    from scipy.spatial.distance import jensenshannon

    for i in range(1, js_dist.shape[0]):
        for j in range(i, js_dist.shape[1]):
            js_dist[i, j] = jensenshannon(X[y == uniq_y[i]], X[y == uniq_y[j]])
    js_dist = js_dist + np.triu(js_dist).T
    return js_dist


def compute_silhouette_score(X, y, metric="euclidean", sample_size=None, **kwargs):
    rng = np.random.default_rng()

    if sample_size is not None:
        indices = rng.permutation(X.shape[0])[:sample_size]
        if metric == "precomputed":
            X, y = X[indices].T[indices].T, y[indices]
        else:
            X, y = X[indices], y[indices]

    return np.mean(metrics.silhouette_samples(X, y, metric=metric, **kwargs))


def compute_lagged_bivariate_prob(y_binary, min_lag, max_lag):
    num_uniq_y = y_binary.shape[1]
    lagged_bivar_prob = np.zeros((max_lag - min_lag + 1, num_uniq_y, num_uniq_y))
    for lag in range(min_lag, max_lag + 1):
        for y_i in range(num_uniq_y):
            for y_j in range(num_uniq_y):
                lagged_bivar_prob[lag - min_lag, y_i, y_j] = np.mean(
                    y_binary[:-lag, y_i] * y_binary[lag:, y_j]
                )
    return lagged_bivar_prob


def cramers_serial_dependency(y_binary, min_lag, max_lag):
    assert y_binary.shape > 1
    hat_pi = np.mean(y_binary, axis=0)
    hat_ij_pi = hat_pi[:, np.newaxis] @ hat_pi[np.newaxis, :]
    cramers = np.zeros(max_lag + 1)
    lagged_bivar_prob = compute_lagged_bivariate_prob(y_binary, min_lag, max_lag)
    for lag in range(min_lag, max_lag + 1):
        cramers[lag] = np.sqrt(
            np.sum((lagged_bivar_prob[lag - min_lag] - hat_ij_pi) ** 2 / hat_ij_pi)
            / (y_binary.shape[1] - 1)
        )
    return cramers


def cohens_serial_dependency(y_binary, min_lag, max_lag):
    hat_pi = np.mean(y_binary, axis=0)
    denom = 1 - np.sum(hat_pi**2)
    cohens = np.zeros(max_lag + 1)
    lagged_bivar_prob = compute_lagged_bivariate_prob(y_binary, min_lag, max_lag)
    for lag in range(min_lag, max_lag + 1):
        cohens[lag] = (
            np.sum(np.diag(lagged_bivar_prob[lag - min_lag]) - hat_pi**2) / denom
        )
    return cohens
